/*  SWAProfile.java -- Software agent profile class

    This is part of SecureFW project, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/dpthesis/index.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

package com.svenda.SecureAlg;
import javacard.framework.*;

public class SWAProfile {
  public byte             m_uuidSWA[];
  public byte             m_encryptKeyAuth[];
  public byte             m_decryptKeyAuth[];
  public byte             m_encryptKeyTransp[];
  public byte             m_decryptKeyTransp[];
  public byte             m_sessionKey[];
  public byte             m_randomNSWA[];
  public byte             m_randomNHWT[];
  public byte             m_authStep;                 // AUTH PROTOCOL STEP
  public boolean          m_authSession;              // TRUE ... SWA IS AUTHENTICATED IN THIS SESSION
  public SecureAlgProfile m_permitedAlgs[];            // USAGE PERMITED ALGORITHMS

  public SWAProfile() {
    m_uuidSWA = new byte[SecureAlgManager.SWA_UUID_LENGTH];
    m_encryptKeyAuth = new byte[SecureAlgManager.AUTH_KEY_LENGTH];
    m_decryptKeyAuth = new byte[SecureAlgManager.AUTH_KEY_LENGTH];
    m_encryptKeyTransp = new byte[SecureAlgManager.TRANSPORT_KEY_LENGTH];
    m_decryptKeyTransp = new byte[SecureAlgManager.TRANSPORT_KEY_LENGTH];
    m_sessionKey = new byte[SecureAlgManager.SESSION_KEY_LENGTH];
    m_randomNSWA = new byte[SecureAlgManager.RANDOM_NONCE_LENGTH];
    m_randomNHWT = new byte[SecureAlgManager.RANDOM_NONCE_LENGTH];
    m_authSession = false;
    m_authStep = SecureAlgManager.SEAUT_MESSAGE_1;
    m_permitedAlgs = new SecureAlgProfile[SecureAlgManager.MAX_ALG_PERMITED_SWA];
  }

  /**
   * Check if this SWAProfile UUID is same as given UUID
   * @param uuidBuffer
   * @param uuidOffset
   * @param uuidLen
   * @return true if given UUID is UUID of this SWA profile, false otherwise
   */
  public boolean IsSWAUUID(byte[] uuidBuffer, short uuidOffset, short uuidLen) {
    boolean status = false;
    if (m_uuidSWA != null) {
      if (Util.arrayCompare(m_uuidSWA, (short) 0, uuidBuffer, uuidOffset, uuidLen) == 0) status = true;
    }
    return status;
  }

  /**
   * Reset security related resources
   */
  public void Reset() {
    m_authSession = false;
    m_authStep = SecureAlgManager.SEAUT_MESSAGE_1;
  }

  /**
   * Free allocated resources
   */
  public void FreeResources() {
    Util.arrayFillNonAtomic(m_uuidSWA, (short) 0, (short) m_uuidSWA.length, (byte) 0);
    Util.arrayFillNonAtomic(m_encryptKeyAuth, (short) 0, (short) m_encryptKeyAuth.length, (byte) 0);
    Util.arrayFillNonAtomic(m_decryptKeyAuth, (short) 0, (short) m_decryptKeyAuth.length, (byte) 0);
    Util.arrayFillNonAtomic(m_encryptKeyTransp, (short) 0, (short) m_encryptKeyAuth.length, (byte) 0);
    Util.arrayFillNonAtomic(m_decryptKeyTransp, (short) 0, (short) m_decryptKeyAuth.length, (byte) 0);
    Util.arrayFillNonAtomic(m_sessionKey, (short) 0, (short) m_sessionKey.length, (byte) 0);
    Util.arrayFillNonAtomic(m_randomNSWA, (short) 0, (short) m_randomNSWA.length, (byte) 0);
    Util.arrayFillNonAtomic(m_randomNHWT, (short) 0, (short) m_randomNHWT.length, (byte) 0);
    m_uuidSWA = null;
    m_encryptKeyAuth = null;
    m_decryptKeyAuth = null;
    m_encryptKeyTransp = null;
    m_decryptKeyTransp = null;
    m_sessionKey = null;
    m_randomNHWT = null;
    m_randomNSWA = null;
    m_permitedAlgs = null;
  }

  /**
   * Add new Algorithm Profile with permited usage
   * @param algProfile - new Algorithm Profile with permited usage
   */
  public void AddPermitedAlg(SecureAlgProfile algProfile) {
    for (byte i = 0; i < m_permitedAlgs.length; i++) {
      if (m_permitedAlgs[i] == null) {
        // OK, FREE SLOT FOUND
        m_permitedAlgs[i] = algProfile;
        return;
      }
    }
    // NO AVAILABLE FREE SLOTS FOR PERMITED SWA
    ISOException.throwIt(SecureAlgManager.SW_SWA_PERMIT_ALG_FULL);
  }
  /**
   * Remove given Algorithm Profile, if is in list of permited Algorithms
   * @param algProfile - Algorithm Profile to remove
   * @return true if Algorithm Profile removed, false otherwise
   */
  public boolean RemoveIfPermitedAlg(byte uuidBuffer[], short uuidOffset, short uuidLen) {
    for (byte i = 0; i < m_permitedAlgs.length; i++) {
      if (m_permitedAlgs[i] != null) {
        if (m_permitedAlgs[i].IsAlgorithm(uuidBuffer, uuidOffset, uuidLen)) {
          m_permitedAlgs[i] = null;
          // NO return, CAUSE REMOVING ALSO POSSIBLE DUPLICITIES
        }
      }
    }

    return false;
  }

  /**
   * Check if algorithm by given UUID is in list of usage permited algorithms
   * @param uuidAlgBuffer
   * @param uuidAlgOffset
   * @param uuidAlgLen
   * @return true if Algorithm usage is permited, false otherwise
   */
  public boolean IsPermitedAlg(byte uuidAlgBuffer[], short uuidAlgOffset, short uuidAlgLen) {
    for (byte i = 0; i < m_permitedAlgs.length; i++) {
      if (m_permitedAlgs[i] != null) {
        if (m_permitedAlgs[i].IsAlgorithm(uuidAlgBuffer, uuidAlgOffset, uuidAlgLen) == true) return true;
      }
    }

    return false;
  }
}