/*  SecureAlgManager.java -- JavaCard applet, base APDU communication class

    This is part of SecureFW project, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/dpthesis/index.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

package com.svenda.SecureAlg;
import javacard.framework.*;
import javacard.security.*;
import javacardx.crypto.*;

public class SecureAlgManager extends javacard.framework.Applet
{
    final static byte DEBUG_MODE_INSTALL_FLAG           = (byte) 0x13;

    final static byte INS_CARD_PROCESS_ALG_DATA		= (byte) 0x03;
    final static byte INS_CARD_PROCESS_MASTER_LICENCE   = (byte) 0x0A;
    final static byte INS_CARD_PROCESS_ALG_LICENCE      = (byte) 0x0B;
    final static byte INS_CARD_SEAUT                    = (byte) 0x30;
    final static byte INS_CARD_SEAUT_STOP               = (byte) 0x31;

    // RESET ALL INTERNAL APPLET VALUES, AVAILABLE ONLY IN DEBUG MODE
    final static byte INS_CARD_DEBUG_RESET              = (byte) 0x40;

    final static byte P1_LICENCE_OUTOFORDER             = (byte) 0xF0;

    // IF APPLET IS INSTALED IN DEBUG MODE, APDU p2 VALUE CAN BE USED TO MODIFY BEHAVIOUR
    // APDU DEBUG (BIT) FLAGS
    final static byte FLAG_DEBUG_DISABLE_IOBUFFER_SECURING    = (byte) 0x01;

    //
    // BASIC CONSTANTS
    //
    final static byte MAX_ALG_PERMITED_SWA              = (byte) 0x08;      // MAXIMUM TOTAL COUNT OF ALGORITHMS, WHICH CAN BE PERMITED FOR ONE SWA
    final static byte MAX_SWA_PROFILES                  = (byte) 0x08;      // MAXIMUM TOTAL COUNT OF SOFTWARE AGENT PROFILES
    final static byte MAX_ALG_PROFILES                  = (byte) 0x08;      // MAXIMUM TOTAL COUNT OF SECURED ALGORITHMS PROFILES

    final static byte ALGORITHM_UUID_LENGTH             = (byte) 0x10;
    final static byte HWT_UUID_LENGTH                   = (byte) 0x10;
    final static byte SWA_UUID_LENGTH                   = (byte) 0x10;
    final static byte RANDOM_NONCE_LENGTH               = (byte) 0x08;
    final static byte AUTH_KEY_LENGTH                   = (byte) 0x10;
    final static byte TRANSPORT_KEY_LENGTH              = (byte) 0x10;
    final static byte SESSION_KEY_LENGTH                = (byte) 0x10;
    final static byte MESSAGE_STOP_LENGTH               = (byte) 0x04;

    // SEAUT PROTOCOL CONSTANTS
    final static byte SEAUT_MESSAGE_1                   = (byte) 0x01;
    final static byte SEAUT_MESSAGE_3                   = (byte) 0x03;


    // ERROR RETURN CODES (HEXADECIMAL CHARACTERS RESERVED FOR FUTURE USE)
    final static short SW_SECURITY_NOT_SATISFIED        = (short) 0x6680;   // SECURITY REQUIRENMENTS NOT SATIFIED
    final static short SW_USAGE_UPDATE_FAIL	        = (short) 0x6681;   // FAIL TO UPDATE ALGORITHM USAGE COUNTER
    final static short SW_USAGE_NOT_ALLOWED	        = (short) 0x6682;   // ALGORITHM USAGE COUNTER DEPLEATED
    final static short SW_ALG_UUID_NOT_UNIQUE           = (short) 0x6683;   // ALGORITHM WITH GIVEN UUID ALREADY EXISTS
    final static short SW_ALG_CONTAINER_FULL            = (short) 0x6684;   // NO MORE ALGORITHMS CAN BE CREATED
    final static short SW_ALG_UUID_NOT_FOUND            = (short) 0x6685;   // ALGORITHM WITH GIVEN UUID NOT EXISTS
    final static short SW_SWA_UUID_NOT_UNIQUE           = (short) 0x6686;   // SOFTWARE AGENT'S PROFILE WITH GIVEN UUID ALREADY EXISTS
    final static short SW_SWA_CONTAINER_FULL            = (short) 0x6687;   // NO MORE SOFTWARE AGENT PROFILES CAN BE CREATED
    final static short SW_BAD_KEY                       = (short) 0x6688;   // INCORRECT KEY
    final static short SW_ALG_TYPE_UNKNOWN              = (short) 0x6689;   // UNKNOWN TYPE OF NEW ALGORITHM (GIVEN IN MASTER LICENCE)
    final static short SW_ALG_UUID_BAD_LENGTH           = (short) 0x6690;   // BAD LENGTH OF ALGORITHM UUID
    final static short SW_SWA_UUID_BAD_LENGTH           = (short) 0x6691;   // BAD LENGTH OF SOFTWARE AGENT PROFILE UUID
    final static short SW_HWT_UUID_BAD                  = (short) 0x6692;   // GIVEN HARDWARE TOKEN UUID IS INCORRECT
    final static short SW_SWA_UUID_BAD                  = (short) 0x6693;   // GIVEN SOFTWARE AGENT UUID IS INCORRECT
    final static short SW_XML_UNEXPECTED_ELEM           = (short) 0x6694;   // UNEXPECTED XML ELEMENT FOUND DURING LICENCE PARSING
    final static short SW_XML_UNEXPECTED_END            = (short) 0x6695;   // UNEXPECTED END OF XML ELEMENT FOUND DURING LICENCE PARSING
    final static short SW_LICENCE_OUTDATED              = (short) 0x6696;   // LICENCE WITH GIVEN ORDINARY NUMBER IS OUTDATED
    final static short SW_SWA_PERMIT_ALG_FULL           = (short) 0x6697;   // NO AVAILABLE FREE SLOTS FOR SWA WHICH CAN BE PERMITED TO ALGORITHM
    final static short SW_ALG_SWA_NOT_PERMITED          = (short) 0x6698;   // USAGE OF ALGORITHM IS NOT PERMITED FOR GIVEN SOFTWARE AGENT
    final static short SW_LICENCE_OUTOFORDER            = (short) 0x6699;   // GIVEN LICENCE IS OUT OF ORDER (NOT DIRECT SUCCESSOR OF LAST ACCEPTED LICENCE), USE SPECIAL FLAG TO PARSE THIS LICENCE
                                                        // 0x6700 RESERVED ISO7816 ERROR
    final static short SW_SWA_AUTH_REQUIRED             = (short) 0x6701;   // AUTHENTIZATION OF SWA IS REQUIRED FOR REQUESTED OPERATION
    final static short SW_ENCRYPT_FAIL                  = (short) 0x6702;   // FAIL TO COMPLETE ENCRYPTION OPERATION
    final static short SW_DECRYPT_FAIL                  = (short) 0x6703;   // FAIL TO COMPLETE DECRYPTION OPERATION
    final static short SW_SWA_NOT_FOUND                 = (short) 0x6704;   // FAIL TO FOUND SOFTWARE AGENT PROFILE BY GIVEN UUID
    final static short SW_ALG_NOT_FOUND                 = (short) 0x6705;   // FAIL TO FOUND ALGORITHM BY GIVEN UUID
    final static short SW_XMLPARSER_INSUFFICIENT_DATA   = (short) 0x6706;   // NOT ALL REQUIRED INFORMATION OBTAINED WHEN END OF LICENCE REACHED
    final static short SW_XMLPARSER_INVALID_DATA        = (short) 0x6707;   // INVALID DATA FORMAT FOUND DURING LICENCE PARSING
    final static short SW_XMLPARSER_UNKNOWN_ERROR       = (short) 0x6708;   // UNKNOWN ERROR OCCURED DURING LICENCE PARSING
    final static short SW_IV_BAD                        = (short) 0x6709;   // BAD INICIALIZATION VECTOR
    final static short SW_CIPHER_DATA_LENGTH_BAD        = (short) 0x6710;   // BAD LENGTH OF DATA USED DURING CIPHER OPERATION
    final static short SW_MESSAGE_LENGTH_BAD            = (short) 0x6711;   // BAD LENGTH OF MESSAGE OBTAINED DURING PROTOCOl (SEAUT)
    final static short SW_SEAUT_STEP_BAD                = (short) 0x6712;   // INCORRECT MESSAGE ORDER OF AUTHENTICATION PART OF SEAUT
    final static short SW_SEAUT_FRESHNESS_FAIL          = (short) 0x6713;   // MESSAGE RECIEVED DURING SEAUT PROTOCOL IS NOT FRESH, ACCIDENTAL DAMAGE OF DATA OR REPLAY ATTACK DETECTED
    final static short SW_SEAUT_STOP_MESSAGE_BAD        = (short) 0x6714;   // BAD STOP MESSAGE USED TO FINALIZE SEAUT PROTOCOL
    final static short SW_KEY_LENGTH_BAD                = (short) 0x6715;   // BAD KEY LENGTH
    final static short SW_ALG_INPUT_BAD                 = (short) 0x6716;   // FORMAT OF INPUT DATA FOR ALGORITHM IS BAD

    SecureAlgProfile                m_secureAlgProfiles[];                  // PROTECTED ALGORITHMS PRIFILES
    SWAProfile                      m_SWAProfiles[];                        // SOFTWARE AGENTS PROFILES
    private byte                    m_masterLicenceKey[];                   // MASTER LICENCE KEY VALUE
    private byte                    m_algLicenceKey[];                      // ALGORITHM KEY VALUE

    private MessageDigest           m_md5Hash;                              // MD5 HASH
    private RandomData              m_rand;                                 // RANDOM NUMBER GENERATOR

    XMLParser                       m_XMLParser;                            // XML PARSER
    XMLMasterLicenceParseHandler    m_XMLMasterLicenceParseHandler;
    XMLAlgLicenceParseHandler       m_XMLAlgLicenceParseHandler;
    private short                   m_lastUsedMasterLicenceOrdNum;          // ORDINARY NUMBER OF LAST ACCEPTED MASTER LICENCE

    JavaCardAES                     m_aesCipher;                            // GENERAL USAGE AES CIPHER

    private byte                    m_tempbuf[];                            // TEMPORARY USAGE BUFFER ALLOCATED IN RAM (FAST)
    private boolean                 m_DEBUGMODE = false;

    private byte                    m_HWT_UUID[];                           // HARDWARE TOKEN UUID, PERSONALIZED DURING APPLET INSTALATION

    // INICIALIZATION VECTOR VALUE, USED FOR CBC MODE
    static byte IV[] = {(byte) 0x2b, (byte) 0x7e, (byte) 0x15, (byte) 0x16,
      (byte) 0x28, (byte) 0xae, (byte) 0xd2, (byte) 0xa6, (byte) 0xab, (byte) 0xf7,
      (byte) 0x15, (byte) 0x88, (byte) 0x09, (byte) 0xcf, (byte) 0x4f, (byte) 0x3c
    };

    // SEAUT CHANNEL DESTROY MESSAGE "STOP"
    final static byte MESSAGE_STOP[] = {(byte) 0x53, (byte) 0x54, (byte) 0x4F, (byte) 0x50};

     /**
     * SecureAlgManager default constructor
     * Only this class's install method should create the applet object.
     */
    protected SecureAlgManager(byte[] buffer, short offset, byte length)
    {
        // data offset is used for application specific parameter.
        // initialization with default offset (AID offset).
        boolean isOP2 = false;

        if(length > 9) {
            // Install parameter detail. Compliant with OP 2.0.1.

            // | size | content
            // |------|---------------------------
            // |  1   | [AID_Length]
            // | 5-16 | [AID_Bytes]
            // |  1   | [Privilege_Length]
            // | 1-n  | [Privilege_Bytes] (normally 1Byte)
            // |  1   | [Application_Proprietary_Length]
            // | 0-m  | [Application_Proprietary_Bytes]

            // APPLICATION PROPRIETARY BYTES
            // | size | content
            // |------|---------------------------
            // |  1   | [DEBUG_MODE_INSTALL_FLAG]
            // | 2-17 | [HWT_UUID]
            // |18-33 | [MASTER LICENCE KEY]


            byte    i;
            byte    instLen = (byte)buffer[offset];
            short   instOffset = (short)(offset+1);
            byte    pri = (byte)buffer[(short)(offset+1+instLen+1)];
            short   dataLen = (short)buffer[(short)(offset+1+instLen+2)];
            short   dataOffset = (short)(offset+1+instLen+3);

            if ((pri & 0xF9) != 0) ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);

            // IF FIRST USER INSTALL DATA VALUE IS EQUAL TO '0x13', THAN ENABLE TO DEBUG MODE
            // DEBUG MODE ALLOW CUT OFF SEAUT BUFFER SECURING
            // !!! DON'T USE THIS IN RELEASE VERSION (SECURITY HOLE)
            if (buffer[dataOffset] == DEBUG_MODE_INSTALL_FLAG) m_DEBUGMODE = true;
            dataOffset++;

            // HARDWARE TOKEN UUID
            m_HWT_UUID = new byte[HWT_UUID_LENGTH];
            Util.arrayCopyNonAtomic(buffer, dataOffset, m_HWT_UUID, (short) 0, HWT_UUID_LENGTH);
            dataOffset += HWT_UUID_LENGTH;

            // SECURED ALGORITHM PROFILES
            m_secureAlgProfiles = new SecureAlgProfile[MAX_ALG_PROFILES];
/*
            // TEST:
            m_secureAlgProfiles[0] = new GaussEaster(buffer, dataOffset, ALGORITHM_UUID_LENGTH);
            dataOffset += ALGORITHM_UUID_LENGTH;
/**/
            // SOFTWARE AGENTS PROFILES
            m_SWAProfiles = new SWAProfile[MAX_SWA_PROFILES];
/*
            // TEST:
            m_SWAProfiles[0] = new SWAProfile();
  	    for (i = 0; i < m_SWAProfiles[0].m_uuidSWA.length; i++) m_SWAProfiles[0].m_uuidSWA[i] = i;
            m_SWAProfiles[0].AddPermitedAlg(m_secureAlgProfiles[0]);
            Util.arrayFillNonAtomic(m_SWAProfiles[0].m_encryptKeyAuth, (short) 0, (short) m_SWAProfiles[0].m_encryptKeyAuth.length, (byte) 0x01);
            Util.arrayFillNonAtomic(m_SWAProfiles[0].m_decryptKeyAuth, (short) 0, (short) m_SWAProfiles[0].m_decryptKeyAuth.length, (byte) 0x02);
            Util.arrayFillNonAtomic(m_SWAProfiles[0].m_encryptKeyTransp, (short) 0, (short) m_SWAProfiles[0].m_encryptKeyTransp.length, (byte) 0x03);
            Util.arrayFillNonAtomic(m_SWAProfiles[0].m_decryptKeyTransp, (short) 0, (short) m_SWAProfiles[0].m_decryptKeyTransp.length, (byte) 0x04);
/**/
            // RANDOM GENERATOR
            m_rand = RandomData.getInstance(RandomData.ALG_SECURE_RANDOM);

            m_masterLicenceKey = new byte[SESSION_KEY_LENGTH];
            m_algLicenceKey = new byte[SESSION_KEY_LENGTH];

            // SET VALUE OF MASTER LICENCE KEY
            Util.arrayCopyNonAtomic(buffer, dataOffset, m_masterLicenceKey, (short) 0, SESSION_KEY_LENGTH);
            dataOffset += SESSION_KEY_LENGTH;

            // XML PARSER INICIALIZATION
            m_XMLParser = new XMLParser();
            m_XMLMasterLicenceParseHandler = new XMLMasterLicenceParseHandler();
            m_XMLAlgLicenceParseHandler = new XMLAlgLicenceParseHandler();

            // MASTER LICENCE WITH ORDINARY NUMBER '1' EXPECTED
            m_lastUsedMasterLicenceOrdNum = (short) 0;

            // AES CIPHER INICIALIZATION
            // TODO: REPLACE WITH javacardx.crypto.cipher(ALG_AES_BLOCK_128_CBC_NOPAD) WHEN HARDWARE AES IS AVALABLE
            m_aesCipher = new JavaCardAES();
/*
            // KEY SCHEDULE FOR SEAUT SESSION KEY ACTUALIZATION
            // PRE-SHARED IV IS USED
            // TODO: IV CAN BE REPLACED BY SOFTWARE AGENT SPECIFIC SESSION KEY ACTUALIZATION VALUE
            m_sessionKeyHash.RoundKeysSchedule(IV, (short) 0);
/**/
            // MD5 HASH ALGORITHM INICIALIZATION
            m_md5Hash = MessageDigest.getInstance(MessageDigest.ALG_MD5, false);

            // TEMPORARY BUFFER USED FOR FAST OPERATION WITH MEMORY
            m_tempbuf = JCSystem.makeTransientByteArray((short) 256, JCSystem.CLEAR_ON_DESELECT);
       } else {
           // <IF NECESSARY, USE COMMENTS TO CHECK LENGTH >
           // if(length != <PUT YOUR PARAMETERS LENGTH> )
           //     ISOException.throwIt((short)(ISO7816.SW_WRONG_LENGTH + length));
       }

        // register this instance
        if (isOP2)
          register(buffer, (short)(offset + 1), (byte)buffer[offset]);
        else
          register();
    }

    /**
     * Method installing the applet.
     * @param bArray the array constaining installation parameters
     * @param bOffset the starting offset in bArray
     * @param bLength the length in bytes of the data parameter in bArray
     */
    public static void install(byte[] bArray, short bOffset, byte bLength) throws ISOException
    {
        /* applet  instance creation */
        new SecureAlgManager (bArray, bOffset, (byte)bLength );
    }

    /**
     * Select method returns true if applet selection is supported.
     * @return boolean status of selection.
     */
    public boolean select()
    {
        ResetSecurityResources();
        // return status of selection
        return true;
    }

    /**
     * Deselect method called by the system in the deselection process.
     */
    public void deselect()
    {
        ResetSecurityResources();

        return;
    }

    /**
     * Method processing an incoming APDU.
     * @see APDU
     * @param apdu the incoming APDU
     * @exception ISOException with the response bytes defined by ISO 7816-4
     */
    public void process(APDU apdu) throws ISOException
    {
        // get the APDU buffer
        byte[] apduBuffer = apdu.getBuffer();

        // ignore the applet select command dispached to the process
        if (selectingApplet())
            return;

        // APDU instruction parser
        switch ( apduBuffer[ISO7816.OFFSET_INS] )
        {
	    case INS_CARD_PROCESS_ALG_DATA:	  ProcessAlgData(apdu); break;
            case INS_CARD_PROCESS_MASTER_LICENCE: ProcessMasterLicence(apdu); break;
            case INS_CARD_PROCESS_ALG_LICENCE:    ProcessAlgLicence(apdu); break;
            case INS_CARD_SEAUT:                  ProcessSEAUT(apdu); break;
            case INS_CARD_SEAUT_STOP:             StopSEAUT(apdu); break;
            case INS_CARD_DEBUG_RESET:            DebugReset(apdu); break;
            default :
                // The INS code is not supported by the dispatcher
                ISOException.throwIt( ISO7816.SW_INS_NOT_SUPPORTED ) ;
            break ;

        }
    }

    //
    //  PRIVATE METHODS
    //

    /**
     * Actualize given session key value, using hash function constructed from block cipher (MATYAS-MEYER-OSEAS)
     * @param sessionKey
     * @param sessionKeyOffset
     * @param sessionKeyLength
     */
    private void ActualizeSEAUTSessionKey(byte sessionKey[], short sessionKeyOffset, short sessionKeyLength) {
      // USING MD5 HASH
      m_md5Hash.reset();
      m_md5Hash.doFinal(sessionKey, sessionKeyOffset, sessionKeyLength, sessionKey, sessionKeyOffset);
/*
      // USING MATYAS-MEYER-OSEAS HASH FROM CIPHER METHOD
      // NOTE: SAME VALUES OF SESSION_KEY_LENGTH AND JavaCardAES.BLOCKLEN IS REQUIRED AND EXPECTED
      short   i;
      short   j;
      short   hlp;
      byte    buf[] = new byte[SESSION_KEY_LENGTH];
      if (sessionKeyLength != JavaCardAES.BLOCKLEN) ISOException.throwIt(SW_KEY_LENGTH_BAD);

      Util.arrayCopyNonAtomic(sessionKey, sessionKeyOffset, buf, (short) 0, sessionKeyLength);
      // NOTE: ONLY ONE BLOCK IS HASHED => KEY IS EVERYTIME SAME (IV)
      JavaCardAES.RoundKeysSchedule(IV, (short) 0);
      JavaCardAES.AESEncryptBlock(sessionKey, sessionKeyOffset);
      hlp = (short) (sessionKeyOffset + sessionKeyLength);
      for (i = sessionKeyOffset, j = 0; i < hlp; i++, j++) sessionKey[i] ^= buf[j];
/**/
    }

    /**
     * Secure outgoing response as defined in SEAUT
     * @param responseBuf
     * @param responseBufOffset
     * @param responseBufLength
     * @param swaProfile
     * @return
     */
    private short SecureResponse(byte responseBuf[], short responseBufOffset, short responseBufLength, SWAProfile swaProfile) {
      short dataLen = 0;

      // SESSION KEY ACTUALIZATION
      ActualizeSEAUTSessionKey(swaProfile.m_sessionKey, (short) 0, (short) swaProfile.m_sessionKey.length);

      // ENCRYPT WITH SWA DECRYPT KEY
      try {
        m_aesCipher.RoundKeysSchedule(swaProfile.m_decryptKeyTransp, (short) 0);
        m_aesCipher.m_IV = swaProfile.m_sessionKey;
        m_aesCipher.m_IVOffset = 0;
        dataLen = m_aesCipher.AESEncryptCBC(responseBuf, (short) responseBufOffset, responseBufLength, true, false);
      }
      catch (ISOException e) {
        throw e;
      }
      catch (Exception e) {
        ISOException.throwIt(SW_ENCRYPT_FAIL);
      }

      // REPEATED XOR WITH SESSION KEY
      for (short i = 0; i < dataLen; i++) {
        responseBuf[(short) (i + responseBufOffset)] ^= swaProfile.m_sessionKey[(short) (i % swaProfile.m_sessionKey.length)];
      }
      return dataLen;
    }

    /**
     * Unsecure incoming request as defined in SEAUT
     * @param requestBuf
     * @param requestBufOffset
     * @param requestBufLength
     * @param swaProfile
     * @return
     */
    private short UnsecureRequest(byte requestBuf[], short requestBufOffset, short requestBufLength, SWAProfile swaProfile) {
      short dataLen = 0;

      // SESSION KEY ACTUALIZATION
      ActualizeSEAUTSessionKey(swaProfile.m_sessionKey, (short) 0, (short) swaProfile.m_sessionKey.length);

      // REPEATED XOR WITH SESSION KEY
      for (short i = 0; i < requestBufLength; i++) {
        requestBuf[(short) (i + requestBufOffset)] ^= swaProfile.m_sessionKey[(short) (i % swaProfile.m_sessionKey.length)];
      }

      // DECRYPT WITH SWA ENCRYPT KEY
      try {
        m_aesCipher.RoundKeysSchedule(swaProfile.m_encryptKeyTransp, (short) 0);
        m_aesCipher.m_IV = swaProfile.m_sessionKey;
        m_aesCipher.m_IVOffset = 0;
        dataLen = m_aesCipher.AESDecryptCBC(requestBuf, (short) requestBufOffset, requestBufLength, true, false);
      }
      catch (ISOException e) {
        throw e;
      }
      catch (Exception e) {
        ISOException.throwIt(SW_DECRYPT_FAIL);
      }

      return dataLen;
    }

    /**
     * Unsecure and parse incoming XML Master licence
     * @param apdu
     */
    private void ProcessMasterLicence(APDU apdu) {
      //
      //  INCOMING MASTER LICENCE MESSAGE {UUIDHWT, MasterLicenceKey(masterLicenceText)}
      //

      byte[]    apdubuf = apdu.getBuffer();
      short     dataLen = apdu.setIncomingAndReceive();
      boolean   outOfOrderLicence = false;
      short     licenceOffset = 0;
      short     licenceLen = 0;

      // CHECK IF OUT OF ORDER LICENCE PARSING ALLOWED
      if (apdubuf[ISO7816.OFFSET_P1] == P1_LICENCE_OUTOFORDER) outOfOrderLicence = true;

      // CHECK TARGET UUIDHWT
      if (Util.arrayCompare(apdubuf, ISO7816.OFFSET_CDATA, m_HWT_UUID, (short) 0, HWT_UUID_LENGTH) != 0) ISOException.throwIt(SW_HWT_UUID_BAD);

      licenceOffset = (short) (ISO7816.OFFSET_CDATA + HWT_UUID_LENGTH);
      licenceLen = (short) (dataLen - HWT_UUID_LENGTH);

      // UNSECURE INCOMING APDU BUFFER USING MASTER LICENCE KEY
      if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
        // DEBUG MODE, NO BUFFER SECURING
      }
      else {
        m_aesCipher.RoundKeysSchedule(m_masterLicenceKey, (short) 0);
        m_aesCipher.m_IV = IV;
        m_aesCipher.m_IVOffset = 0;

        licenceLen = m_aesCipher.AESDecryptCBC(apdubuf, licenceOffset, licenceLen, true, true);
      }

      // PARSE LICENCE
      ParseMasterLicence(apdubuf, licenceOffset, licenceLen, outOfOrderLicence);
    }

    /**
     * Unsecure and parse incoming XML Algorithm licence
     * @param apdu
     */
    private void ProcessAlgLicence(APDU apdu) {
      //
      //  INCOMING ALGORITHM LICENCE MESSAGE {UUIDHWT, AlgLicenceKey(algLicenceText)}
      //

      // UNSECURE INCOMING APDU BUFFER USING MASTER LICENCE KEY
      byte[]    apdubuf = apdu.getBuffer();
      short     dataLen = apdu.setIncomingAndReceive();
      boolean   outOfOrderLicence = false;
      short     licenceOffset = 0;
      short     licenceLen = 0;

      // CHECK IF OUT OF ORDER LICENCE PARSING ALLOWED
      if (apdubuf[ISO7816.OFFSET_P1] == P1_LICENCE_OUTOFORDER) outOfOrderLicence = true;

      licenceOffset = (short) (ISO7816.OFFSET_CDATA + HWT_UUID_LENGTH);
      licenceLen = (short) (dataLen - HWT_UUID_LENGTH);

      if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
        // DEBUG MODE, NO BUFFER SECURING
      }
      else {
        m_aesCipher.RoundKeysSchedule(m_algLicenceKey, (short) 0);
        m_aesCipher.m_IV = IV;
        m_aesCipher.m_IVOffset = 0;
        licenceLen = m_aesCipher.AESDecryptCBC(apdubuf, licenceOffset, licenceLen, true, true);
      }

      // CHECK TARGET UUIDHWT
      if (Util.arrayCompare(apdubuf, ISO7816.OFFSET_CDATA, m_HWT_UUID, (short) 0, HWT_UUID_LENGTH) != 0) ISOException.throwIt(SW_HWT_UUID_BAD);

      // PARSE LICENCE
      ParseAlgLicence(apdubuf, licenceOffset, licenceLen, outOfOrderLicence);
    }

    private void ParseMasterLicence(byte licenceBuffer[], short licenceOffset, short licenceLen, boolean outOfOrderLicence) {
      // PARSE LICENCE
      try {
        m_XMLMasterLicenceParseHandler.m_secureAlgProfiles = m_secureAlgProfiles;
        m_XMLMasterLicenceParseHandler.m_SWAProfiles = m_SWAProfiles;
        m_XMLMasterLicenceParseHandler.m_uuidHWT = m_HWT_UUID;
        m_XMLMasterLicenceParseHandler.m_lastUsedLicenceOrdNum = m_lastUsedMasterLicenceOrdNum;
        m_XMLMasterLicenceParseHandler.m_outOfOrderLicence = outOfOrderLicence;

        m_XMLParser.parse(m_XMLMasterLicenceParseHandler, licenceBuffer, licenceOffset, licenceLen);
        // APPLY VALUES FROM LICENCE
        if (m_XMLMasterLicenceParseHandler.m_correctLicence) {
          // LICENCE ORD NUM
          m_lastUsedMasterLicenceOrdNum = m_XMLMasterLicenceParseHandler.m_lastUsedLicenceOrdNum;
          // MASTER LICENCE KEY
          if (m_XMLMasterLicenceParseHandler.m_masterLicenceKey != null) {
            Util.arrayCopyNonAtomic(m_XMLMasterLicenceParseHandler.m_masterLicenceKey, (short) 0, m_masterLicenceKey, (short) 0, (short) m_XMLMasterLicenceParseHandler.m_masterLicenceKey.length);
          }
          // ALG LICENCE KEY
          if (m_XMLMasterLicenceParseHandler.m_algLicenceKey != null) {
            Util.arrayCopyNonAtomic(m_XMLMasterLicenceParseHandler.m_algLicenceKey, (short) 0, m_algLicenceKey, (short) 0, (short) m_XMLMasterLicenceParseHandler.m_algLicenceKey.length);
          }
        }
      }
      catch (ISOException e) {
        throw e;
      }
      catch (Exception e) {
        ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_UNKNOWN_ERROR);
      }
    }

     private void ParseAlgLicence(byte licenceBuffer[], short licenceOffset, short licenceLen, boolean outOfOrderLicence) {
      // PARSE LICENCE
      try {
        m_XMLAlgLicenceParseHandler.m_secureAlgProfiles = m_secureAlgProfiles;
        m_XMLAlgLicenceParseHandler.m_actualUUIDHWT = m_HWT_UUID;
        m_XMLAlgLicenceParseHandler.m_outOfOrderLicence = outOfOrderLicence;

        m_XMLParser.parse(m_XMLAlgLicenceParseHandler, licenceBuffer, licenceOffset, licenceLen);
      }
      catch (ISOException e) {
        throw e;
      }
      catch (Exception e) {
        ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_UNKNOWN_ERROR);
      }
    }

    private void ProcessAlgData(APDU apdu) {
      byte[]            apdubuf = apdu.getBuffer();
      short             dataLen = apdu.setIncomingAndReceive();

      short             dataOffset = 0;
      short             outputDataLen = 0;
      SWAProfile        actualSWA;
      SecureAlgProfile  actualAlg;

      dataOffset = ISO7816.OFFSET_CDATA;

      // CHECK TARGET UUIDHWT
      if (Util.arrayCompare(apdubuf, dataOffset, m_HWT_UUID, (short) 0, HWT_UUID_LENGTH) != 0) ISOException.throwIt(SW_HWT_UUID_BAD);
      dataOffset += HWT_UUID_LENGTH;
      dataLen -= HWT_UUID_LENGTH;

      // GET SWA PROFILE FOR GIVEN UUIDSWA
      if ((actualSWA = GetSWAByUUID(m_SWAProfiles, apdubuf, dataOffset, SWA_UUID_LENGTH)) == null) ISOException.throwIt(SW_SWA_UUID_BAD);
      dataOffset += SWA_UUID_LENGTH;
      dataLen -= SWA_UUID_LENGTH;

      // CHECK AUTHENTIZATION
      if (!actualSWA.m_authSession) ISOException.throwIt(SW_SWA_AUTH_REQUIRED);

      // UNSECURE INCOMING REQUEST
      if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
        // DEBUG MODE, NO BUFFER SECURING
      }
      else dataLen = UnsecureRequest(apdubuf, dataOffset, dataLen, actualSWA);

      //
      //  PROCESS SECURED ALGORITHM OVER INPUT DATA
      //
      if ((actualAlg = GetAlgorithmByUUID(m_secureAlgProfiles, apdubuf, dataOffset, ALGORITHM_UUID_LENGTH)) == null) ISOException.throwIt(SW_ALG_UUID_NOT_FOUND);

      // CHECK IF ALGORITHM IS ON SWA USAGE PERMITED LIST
      if (!actualSWA.IsPermitedAlg(apdubuf, dataOffset, ALGORITHM_UUID_LENGTH)) ISOException.throwIt(SW_ALG_SWA_NOT_PERMITED);

      dataOffset += ALGORITHM_UUID_LENGTH;
      dataLen -= ALGORITHM_UUID_LENGTH;

      // COPY DATA FOR ALGORITHM INTO WORK BUFFER
      Util.arrayCopyNonAtomic(apdubuf, dataOffset, m_tempbuf, (short) 0, dataLen);

      // PROCESS DATA BY GIVEN ALGORITHM
      dataLen = actualAlg.ProcessAlgData(m_tempbuf, (short) 0, dataLen);

      // SECURE OUTGOING RESPONSE
      if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
        // DEBUG MODE, NO BUFFER SECURING
      }
      else dataLen = SecureResponse(m_tempbuf, (short) 0, dataLen, actualSWA);

      //
      // CREATE OUTGOING MESSAGE {UUIDHWT, UUIDSWA, XKr(decryptKeySWA(output_data))}
      //
      dataOffset = ISO7816.OFFSET_CDATA;
      outputDataLen = 0;
      // UUIDHWT
      Util.arrayCopyNonAtomic(m_HWT_UUID, (short) 0, apdubuf, dataOffset, HWT_UUID_LENGTH);
      dataOffset += HWT_UUID_LENGTH;
      outputDataLen += HWT_UUID_LENGTH;

      // UUIDSWA
      Util.arrayCopyNonAtomic(actualSWA.m_uuidSWA, (short) 0, apdubuf, dataOffset, SWA_UUID_LENGTH);
      dataOffset += SWA_UUID_LENGTH;
      outputDataLen += SWA_UUID_LENGTH;

      // output_data
      Util.arrayCopyNonAtomic(m_tempbuf, (short) 0, apdubuf, dataOffset, dataLen);
      outputDataLen += dataLen;

      // OUTPUT OUTGOING BUFFER
      apdu.setOutgoingAndSend(ISO7816.OFFSET_CDATA, outputDataLen);
    }

    /**
     * Return algorithm profile by given UUID
     * @param secureAlgProfiles
     * @param dataBuffer
     * @param offset
     * @param dataLen
     * @return algorithm profile
     */
    public static SecureAlgProfile GetAlgorithmByUUID(SecureAlgProfile secureAlgProfiles[], byte dataBuffer[], short offset, short dataLen) {
      short i;

      if (dataLen < ALGORITHM_UUID_LENGTH) ISOException.throwIt(SW_ALG_UUID_BAD_LENGTH);
      for (i = 0; i < secureAlgProfiles.length; i++) {
        if (secureAlgProfiles[i] != null) {
          if (secureAlgProfiles[i].IsAlgorithm(dataBuffer, offset, dataLen)) return secureAlgProfiles[i];
        }
      }
      return null;
    }

    /**
     * Reset security related resources to default state
     */
    private void ResetSecurityResources() {
      for (byte i = 0; i < m_SWAProfiles.length; i++) {
        if (m_SWAProfiles[i] != null) m_SWAProfiles[i].Reset();
      }
    }

    /**
     * Return software agent profile by given UUID
     * @param SWAProfiles
     * @param uuidSWA
     * @param uuidSWAOffset
     * @param uuidSWALen
     * @return
     */
    public static SWAProfile GetSWAByUUID(SWAProfile SWAProfiles[], byte[] uuidSWA, short uuidSWAOffset, short uuidSWALen) {
      for (byte i = 0; i < SWAProfiles.length; i++) {
        if (SWAProfiles[i] != null) {
          if (SWAProfiles[i].IsSWAUUID(uuidSWA, uuidSWAOffset, uuidSWALen)) return SWAProfiles[i];
        }
      }

      return null;
    }

    /**
     * Check, if new UUID is not used for already existing algorithm profile
     * @param profiles
     * @param newUUID
     * @return true if unique, false otherwise
     */
    public static boolean IsAlgUUIDUnique(SecureAlgProfile[] profiles, byte[] newUUID) {
      boolean   unique = true;
      // CHECK FOR UNIQUE OF ALGORITHM UUID
      for (byte i = 0; i < profiles.length; i++) {
        if ((newUUID != null) && (profiles[i] != null)) {
          if (profiles[i].IsAlgorithm(newUUID, (short) 0, (short) newUUID.length)) {
            // ALREADY EXISTS
            unique = false;
            break;
          }
        }
      }

      return unique;
    }

    /**
     * Check, if new UUID is not used for already existing software agent profile
     * @param SWAProfiles
     * @param newUUID
     * @return true if unique, false otherwise
     */
    public static boolean IsSWAUUIDUnique(SWAProfile[] SWAProfiles, byte[] newUUID) {
      boolean   unique = true;
      // CHECK FOR UNIQUE OF ALGORITHM UUID
      for (byte i = 0; i < SWAProfiles.length; i++) {
        if ((newUUID != null) && (SWAProfiles[i] != null)) {
          if (SWAProfiles[i].IsSWAUUID(newUUID, (short) 0, (short) newUUID.length)) {
            // ALREADY EXISTS
            unique = false;
            break;
          }
        }
      }

      return unique;
    }

    /**
     * Remove given algorithm (by UUID) from all software agents, if software agent is permited to use this algorithm.
     * Call when algorithm profile is removed.
     * @param SWAProfiles
     * @param uuidBuffer  algorithm UUID
     * @param uuidOffset
     * @param uuidLen
     */
    public static void RemoveAllPermitedAlgsByUUID(SWAProfile SWAProfiles[], byte uuidBuffer[], short uuidOffset, short uuidLen) {
      for (byte i = 0; i < SWAProfiles.length; i++) {
        if (SWAProfiles[i] != null) {
          SWAProfiles[i].RemoveIfPermitedAlg(uuidBuffer, uuidOffset, uuidLen);
        }
      }
    }


    /**
     * Process authentization protocol SEAUT
     * @param apdu Incoming APDU contains SEAUT message
     */
    private void ProcessSEAUT(APDU apdu) {
      boolean     stepProcessCorrect = false;
      short       apdubufOffset = 0;
      short       outputDataLen = 0;
      byte[]      decryptKeySWA;
      SWAProfile     actualSWA = null;
      byte[]      apdubuf = apdu.getBuffer();
      short       apdubufLen = apdu.setIncomingAndReceive();

      // CHECK EXPECTED MESSAGE LENGTH (AT LEAST UUIDHWT AND UUIDSWA EXPECTED)
      if (apdubufLen < (short) (HWT_UUID_LENGTH + SWA_UUID_LENGTH)) ISOException.throwIt(SecureAlgManager.SW_MESSAGE_LENGTH_BAD);

      apdubufOffset = ISO7816.OFFSET_CDATA;
      // CHECK TARGET UUIDHWT
      if (Util.arrayCompare(apdubuf, apdubufOffset, m_HWT_UUID, (short) 0, HWT_UUID_LENGTH) != 0) ISOException.throwIt(SW_HWT_UUID_BAD);
      apdubufOffset += HWT_UUID_LENGTH;

      // OBTAIN SWA PROFILE RESPECT TO UUIDSWA
      if ((actualSWA = GetSWAByUUID(m_SWAProfiles, apdubuf, apdubufOffset, SWA_UUID_LENGTH)) == null) ISOException.throwIt(SW_SWA_UUID_BAD);
      apdubufOffset += SWA_UUID_LENGTH;

      if (actualSWA != null ) {
        // SEAUT MESSAGE MUST BE SUCCESSOR OF PREVIOUS PROCESSED MESSAGE
        if (apdubuf[ISO7816.OFFSET_P1] != actualSWA.m_authStep) ISOException.throwIt(SecureAlgManager.SW_SEAUT_STEP_BAD);

        // PROCESS SEAUT MESSAGE
        switch (apdubuf[ISO7816.OFFSET_P1]) {
          case SEAUT_MESSAGE_1: {
            //
            // INCOMING MESSAGE {UUIDHWT, UUIDSWA, NSWA}
            //

            // CHECK EXPECTED MESSAGE LENGTH
            if (apdubufLen != (short) (HWT_UUID_LENGTH + SWA_UUID_LENGTH + RANDOM_NONCE_LENGTH)) ISOException.throwIt(SecureAlgManager.SW_MESSAGE_LENGTH_BAD);

            // GET BACKUP COPY RANDOM NONCE NSWA
            Util.arrayCopyNonAtomic(apdubuf, apdubufOffset, actualSWA.m_randomNSWA, (short) 0, RANDOM_NONCE_LENGTH);
            apdubufOffset += RANDOM_NONCE_LENGTH;

            // GENERATE RANDOM NONCE NHWT
            m_rand.generateData(actualSWA.m_randomNHWT, (short) 0, RANDOM_NONCE_LENGTH);
            // TEST: SET TO ALL '2' FOR TESTING PURPOSES
            Util.arrayFillNonAtomic(actualSWA.m_randomNHWT, (short) 0, RANDOM_NONCE_LENGTH, (byte) 2);

            //
            //  OUTGOING MESSAGE {UUIDHWT, UUIDSWA, EKDa(NHWT, NSWA, UUIDSWA)}
            //
            apdubufOffset = 0;
            // NHWT
            Util.arrayCopyNonAtomic(actualSWA.m_randomNHWT, (short) 0, m_tempbuf, apdubufOffset, RANDOM_NONCE_LENGTH);
            apdubufOffset += RANDOM_NONCE_LENGTH;
            // NSWA
            Util.arrayCopyNonAtomic(actualSWA.m_randomNSWA, (short) 0, m_tempbuf, apdubufOffset, RANDOM_NONCE_LENGTH);
            apdubufOffset += RANDOM_NONCE_LENGTH;
            // UUIDSWA
            Util.arrayCopyNonAtomic(actualSWA.m_uuidSWA, (short) 0, m_tempbuf, apdubufOffset, SWA_UUID_LENGTH);
            apdubufOffset += SWA_UUID_LENGTH;

            // ENCRYPT (NHWT, NSWA, UUIDSWA) USING KDa
            if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
              // DEBUG MODE, NO BUFFER SECURING
              outputDataLen = apdubufOffset;
            }
            else {
              m_aesCipher.RoundKeysSchedule(actualSWA.m_decryptKeyAuth, (short) 0);
              m_aesCipher.m_IV = IV;
              m_aesCipher.m_IVOffset = 0;
              // NO PADDING, NO STRENGTH LAST BLOCK
              outputDataLen = m_aesCipher.AESEncryptCBC(m_tempbuf, (short) 0, apdubufOffset, false, false);
            }

            apdubufOffset = ISO7816.OFFSET_CDATA;
            // UUIDHWT
            Util.arrayCopyNonAtomic(m_HWT_UUID, (short) 0, apdubuf, apdubufOffset, HWT_UUID_LENGTH);
            apdubufOffset += HWT_UUID_LENGTH;
            // UUIDSWA
            Util.arrayCopyNonAtomic(actualSWA.m_uuidSWA, (short) 0, apdubuf, apdubufOffset, SWA_UUID_LENGTH);
            apdubufOffset += SWA_UUID_LENGTH;
            // EKDa(NHWT, NSWA, UUIDSWA)
            Util.arrayCopyNonAtomic(m_tempbuf, (short) 0, apdubuf, apdubufOffset, outputDataLen);

            // OUTPUT MESSAGE LENGTH (HWT_UUID_LENGTH + SWA_UUID_LENGTH + outputDataLen)
            outputDataLen += HWT_UUID_LENGTH + SWA_UUID_LENGTH;

            // SEND OUTGOING APDU
            apdu.setOutgoingAndSend(ISO7816.OFFSET_CDATA, outputDataLen);

            // SET AUTH STEP TO MESSAGE 3
            actualSWA.m_authStep = SEAUT_MESSAGE_3;

            break;
          }
          case SEAUT_MESSAGE_3: {
            //
            // INCOMING MESSAGE {UUIDHWT, UUIDSWA, EKEa(NSWA, NHWT)}
            //
            if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
              // DEBUG MODE, NO BUFFER SECURING
              outputDataLen = (short) (apdubufLen - (apdubufOffset - ISO7816.OFFSET_CDATA));
            }
            else {
              // DECRYPT EKEa(NSWA, NHWT) USING KEa
              m_aesCipher.RoundKeysSchedule(actualSWA.m_encryptKeyAuth, (short) 0);
              m_aesCipher.m_IV = IV;
              m_aesCipher.m_IVOffset = 0;
              // NO PADDING, NO STRENGTH LAST BLOCK
              outputDataLen = m_aesCipher.AESDecryptCBC(apdubuf, apdubufOffset, (short) (apdubufLen - (apdubufOffset - ISO7816.OFFSET_CDATA)), false, false);
            }

            // CHECK EXPECTED NSWA FRESHNESS USING NHWT
            if (Util.arrayCompare(actualSWA.m_randomNSWA, (short) 0, apdubuf, apdubufOffset, RANDOM_NONCE_LENGTH) != 0) ISOException.throwIt(SecureAlgManager.SW_SEAUT_FRESHNESS_FAIL);
            apdubufOffset += RANDOM_NONCE_LENGTH;
            // CHECK MESSAGE FRESHNESS USING NHWT
            if (Util.arrayCompare(actualSWA.m_randomNHWT, (short) 0, apdubuf, apdubufOffset, RANDOM_NONCE_LENGTH) != 0) ISOException.throwIt(SecureAlgManager.SW_SEAUT_FRESHNESS_FAIL);

            // NOTE: SWA IS AUTHENTICATED AGAINST HWT NOW
            actualSWA.m_authSession = true;
            actualSWA.m_authStep = SEAUT_MESSAGE_1;

            // CONSTRUCT SESSION KEY KR (Hash(NHWT | NSWA | UUIDSWA | NSWA | UUIDSWA))
            apdubufOffset = 0;
            Util.arrayCopyNonAtomic(actualSWA.m_randomNHWT, (short) 0, m_tempbuf, apdubufOffset, RANDOM_NONCE_LENGTH);
            apdubufOffset += RANDOM_NONCE_LENGTH;
            Util.arrayCopyNonAtomic(actualSWA.m_randomNSWA, (short) 0, m_tempbuf, apdubufOffset, RANDOM_NONCE_LENGTH);
            apdubufOffset += RANDOM_NONCE_LENGTH;
            Util.arrayCopyNonAtomic(actualSWA.m_uuidSWA, (short) 0, m_tempbuf, apdubufOffset, SWA_UUID_LENGTH);
            apdubufOffset += SWA_UUID_LENGTH;
            Util.arrayCopyNonAtomic(actualSWA.m_randomNSWA, (short) 0, m_tempbuf, apdubufOffset, RANDOM_NONCE_LENGTH);
            apdubufOffset += RANDOM_NONCE_LENGTH;
            Util.arrayCopyNonAtomic(actualSWA.m_uuidSWA, (short) 0, m_tempbuf, apdubufOffset, SWA_UUID_LENGTH);
            apdubufOffset += SWA_UUID_LENGTH;

            // USING MD5 HASH
            m_md5Hash.reset();
            m_md5Hash.doFinal(m_tempbuf, (short) 0, apdubufOffset, actualSWA.m_sessionKey, (short) 0);

            break;
          }
          default: {
            ISOException.throwIt(SecureAlgManager.SW_SEAUT_STEP_BAD);
          }
        }
      }
    }

    /**
     * Stop SEAUT authentization for given software agent
     * @param apdu
     */
    private void StopSEAUT(APDU apdu) {
      byte[]            apdubuf = apdu.getBuffer();
      short             dataLen = apdu.setIncomingAndReceive();
      short             dataOffset = 0;
      short             outputDataLen = 0;
      SWAProfile        actualSWA;
      SecureAlgProfile  actualAlg;

      dataOffset = ISO7816.OFFSET_CDATA;
      // CHECK TARGET UUIDHWT
      if (Util.arrayCompare(apdubuf, dataOffset, m_HWT_UUID, (short) 0, HWT_UUID_LENGTH) != 0) ISOException.throwIt(SW_HWT_UUID_BAD);
      dataOffset += HWT_UUID_LENGTH;
      dataLen -= HWT_UUID_LENGTH;

      // GET SWA PROFILE FOR GIVEN UUIDSWA
      if ((actualSWA = GetSWAByUUID(m_SWAProfiles, apdubuf, dataOffset, SWA_UUID_LENGTH)) == null) ISOException.throwIt(SW_SWA_UUID_BAD);
      dataOffset += SWA_UUID_LENGTH;
      dataLen -= SWA_UUID_LENGTH;

      // CHECK AUTHENTIZATION
      if (!actualSWA.m_authSession) ISOException.throwIt(SW_SWA_AUTH_REQUIRED);

      // UNSECURE INCOMING REQUEST
      if (m_DEBUGMODE && ((apdubuf[ISO7816.OFFSET_P2] & FLAG_DEBUG_DISABLE_IOBUFFER_SECURING) != 0)) {
        // DEBUG MODE, NO BUFFER SECURING
      }
      else dataLen = UnsecureRequest(apdubuf, dataOffset, dataLen, actualSWA);

      // COMPARE REQUEST TO BE EQUAL TO STOP MESSAGE
      if ((dataLen == MESSAGE_STOP_LENGTH) &&
          (Util.arrayCompare(apdubuf, dataOffset, MESSAGE_STOP, (short) 0, MESSAGE_STOP_LENGTH) == 0)) {

          // STOP SEAUT
          actualSWA.Reset();
      }
      else {
          // BAD STOP MESSAGE
          ISOException.throwIt(SW_SEAUT_STOP_MESSAGE_BAD);
      }
    }

    /**
     * Reset and release internal structures to the state after install()
     * Available only in DEBUG MODE !!!
     * @param apdu
     */
    private void DebugReset(APDU apdu) {
      byte i;

      if (!m_DEBUGMODE) ISOException.throwIt(SW_SECURITY_NOT_SATISFIED);

      // RESET SWA PROFILES
      for (i = 0; i < m_SWAProfiles.length; i++) m_SWAProfiles[i] = null;
      // RESET ALGORITHM PROFILES
      for (i = 0; i < m_secureAlgProfiles.length; i++) m_SWAProfiles[i] = null;

      // MASTER LICENCE WITH ORDINARY NUMBER '1' EXPECTED
      m_lastUsedMasterLicenceOrdNum = (short) 0;
    }
}

