/*  XMLAlgLicenceParseHandler.java -- XML parsing handler for algorithm licences

    This is part of SecureFW project, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/dpthesis/index.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

package com.svenda.SecureAlg;
import javacard.framework.*;

public class XMLAlgLicenceParseHandler implements XMLParseHandler {
  //
  // XML PARSER TAGS
  //
  private final static byte[] XML_TAGS = {
    // \0
    0x00,
    // ALic
    0x41, 0x4C, 0x69, 0x63,
    // UUIDHWT
    0x55, 0x55, 0x49, 0x44, 0x48, 0x57, 0x54,
    // UUID
    0x55, 0x55, 0x49, 0x44,
    // LON - Licence Ordinary Number
    0x4C, 0x4F, 0x4E,
    // UCount
    0x55, 0x43, 0x6F, 0x75, 0x6E, 0x74
  };

  // "ALic"
  private final static byte XML_TAGS_ALic_OFFSET = 1;
  private final static byte XML_TAGS_ALic_LENGTH = 4;
  // "UUIDHWT"
  private final static byte XML_TAGS_UUIDHWT_OFFSET = 5;
  private final static byte XML_TAGS_UUIDHWT_LENGTH = 7;
  // "UUID"
  private final static byte XML_TAGS_UUID_OFFSET = 12;
  private final static byte XML_TAGS_UUID_LENGTH = 4;
  // "LON"
  private final static byte XML_TAGS_LON_OFFSET = 16;
  private final static byte XML_TAGS_LON_LENGTH = 3;
  // "UCount"
  private final static byte XML_TAGS_UCount_OFFSET = 19;
  private final static byte XML_TAGS_UCount_LENGTH = 6;

  private final static byte TAG_ALGLIC            = (byte) 0x01;
  private final static byte TAG_UUIDHWT           = (byte) 0x02;
  private final static byte TAG_UUID              = (byte) 0x03;
  private final static byte TAG_LICORDNUM         = (byte) 0x04;
  private final static byte TAG_UCOUNT            = (byte) 0x05;

  private final static byte TAG_NOTAG             = (byte) 0xFF;

  // LIST OF ALL AVAILABLE PROFILES
  public SecureAlgProfile    m_secureAlgProfiles[];
  // UUID OF ACTUAL HWT
  public byte         m_actualUUIDHWT[];
  public boolean      m_outOfOrderLicence = false;

  private boolean     m_processLicence = false;

  private byte[]      m_UUID = null;
  private byte[]      m_UUIDHWT = null;
  private short       m_licOrdNum = 0;
  private short       m_count = 0;
  private Stack       m_tagStack = new Stack();
  //
  //  METHODS
  //
  public XMLAlgLicenceParseHandler() {}

  /**
   * Pop and return topmost tag from stack
   */
  protected byte popTag() {
    if (m_tagStack.empty()) return TAG_NOTAG;
    else return m_tagStack.pop();
  }

  /**
   * Return topmost tag from stack
   */
  protected byte topTag() {
    if (m_tagStack.empty()) return TAG_NOTAG;
    else return m_tagStack.top();
  }


  // Implementation of DocHandler is below this line
  public void startDocument() {
    m_tagStack.clear();
    m_UUID = null;
    m_UUIDHWT = null;
    m_licOrdNum = 0;
    m_count = 0;
  }
  public void endDocument() {
    if ((m_UUIDHWT == null) || (m_UUID == null)) ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_INSUFFICIENT_DATA);
    // PARSED UUIDHWT MUST BE SAME AS OBTAINED FROM PARENT
    if (Util.arrayCompare(m_UUIDHWT, (short) 0, m_actualUUIDHWT, (short) 0, (short) m_actualUUIDHWT.length) != 0) ISOException.throwIt(SecureAlgManager.SW_HWT_UUID_BAD);

    // FIND TARGET ALGORITHM BY UUID
    byte    i;
    boolean algFound = false;
    for (i = 0; i < m_secureAlgProfiles.length; i++) {
      if ((m_secureAlgProfiles[i] != null) && m_secureAlgProfiles[i].IsAlgorithm(m_UUID, (short) 0, (short) m_UUID.length)) {
        algFound = true;
        // APPLY PARSED VALUES ONLY IF LICENCE ORD NUMBER IS HIGHER THAN LAST ALREADY USED
        if (m_secureAlgProfiles[i].UpdateLastLicenceOrdNum(m_licOrdNum, m_outOfOrderLicence)) {
          // UPDATING VALUES
          m_secureAlgProfiles[i].UpdateUsageCounter(m_count);
        }
        else ISOException.throwIt(SecureAlgManager.SW_LICENCE_OUTDATED);

        break;
      }
    }

    // TEST IF ALGORITHM WAS FOUND
    if (!algFound) ISOException.throwIt(SecureAlgManager.SW_ALG_NOT_FOUND);
  }

  public void startElement(CardString tag, CardString attributName, CardString attributValue, boolean newElement) throws Exception {
    //
    // ELEMENT AlgsLic
    //
    if (tag.compare(XML_TAGS, XML_TAGS_ALic_OFFSET, XML_TAGS_ALic_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_ALGLIC);
      return;
    }
    //
    // ELEMENT UUIDHWT
    //
    if (tag.compare(XML_TAGS, XML_TAGS_UUIDHWT_OFFSET, XML_TAGS_UUIDHWT_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_UUIDHWT);
      return;
    }
    //
    // ELEMENT UUID
    //
    if (tag.compare(XML_TAGS, XML_TAGS_UUID_OFFSET, XML_TAGS_UUID_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_UUID);
      return;
    }
    //
    // ELEMENT UCount
    //
    if (tag.compare(XML_TAGS, XML_TAGS_UCount_OFFSET, XML_TAGS_UCount_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_UCOUNT);
      return;
    }
    //
    // ELEMENT LOC
    //
    if (tag.compare(XML_TAGS, XML_TAGS_LON_OFFSET, XML_TAGS_LON_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_LICORDNUM);
      return;
    }
  }

  public void startElementFinal(CardString tag) {}

  public void endElement(CardString elem) throws Exception{
    //
    // ELEMENT AlgsLic
    //
    if (elem.compare(XML_TAGS, XML_TAGS_ALic_OFFSET, XML_TAGS_ALic_LENGTH) == 0) {
      if (topTag() == TAG_ALGLIC) popTag();
      else throw new Exception();
      return;
    }
    //
    // ELEMENT UUIDHWT
    //
    if (elem.compare(XML_TAGS, XML_TAGS_UUIDHWT_OFFSET, XML_TAGS_UUIDHWT_LENGTH) == 0) {
      if (topTag() == TAG_UUIDHWT) popTag();
      else throw new Exception();
      return;
    }
    //
    // ELEMENT UUID
    //
    if (elem.compare(XML_TAGS, XML_TAGS_UUID_OFFSET, XML_TAGS_UUID_LENGTH) == 0) {
      if (topTag() == TAG_UUID) popTag();
      else throw new Exception();
      return;
    }
    //
    // ELEMENT UCount
    //
    if (elem.compare(XML_TAGS, XML_TAGS_UCount_OFFSET, XML_TAGS_UCount_LENGTH) == 0) {
      if (topTag() == TAG_UCOUNT) popTag();
      else throw new Exception();
      return;
    }
    //
    // ELEMENT LOC
    //
    if (elem.compare(XML_TAGS, XML_TAGS_LON_OFFSET, XML_TAGS_LON_LENGTH) == 0) {
      if (topTag() == TAG_LICORDNUM) popTag();
      else throw new Exception();
      return;
    }
  }

  public void text(CardString text) throws Exception{
    switch (topTag()) {
      case TAG_UUIDHWT: {
        m_UUIDHWT = new byte[text.m_end];
        Util.arrayCopy(text.m_buffer, (short) 0, m_UUIDHWT, (short) 0, text.m_end);
        break;
      }
      case TAG_UUID: {
        m_UUID = new byte[text.m_end];
        Util.arrayCopy(text.m_buffer, (short) 0, m_UUID, (short) 0, text.m_end);
        break;
      }
      case TAG_LICORDNUM: {
        m_licOrdNum = text.getShortNumber();
        break;
      }
      case TAG_UCOUNT: {
        m_count = text.getShortNumber();
        break;
      }
      default: {
        // NO ACTION
      }
    }
  }
  // implementation of DocHandler is above this line
}