/*  XMLMasterLicenceParseHandler.java -- XML parsing handler for master licences

    This is part of SecureFW project, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/dpthesis/index.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

package com.svenda.SecureAlg;
import javacard.framework.*;


public class XMLMasterLicenceParseHandler implements XMLParseHandler {
  //
  // XML PARSER TAGS
  //
  private final static byte[] XML_TAGS = {
    // \0
    0x00,
    // MLic
    0x4D, 0x4C, 0x69, 0x63,
    // UUIDHWT
    0x55, 0x55, 0x49, 0x44, 0x48, 0x57, 0x54,
    // MLicKey
    0x4D, 0x4C, 0x69, 0x63, 0x4B, 0x65, 0x79,
    // ALicKey
    0x41, 0x4C, 0x69, 0x63, 0x4B, 0x65, 0x79,
    // NewAlg
    0x4E, 0x65, 0x77, 0x41, 0x6C, 0x67,
    // ModifyAlg
    0x4D, 0x6F, 0x64, 0x69, 0x66, 0x79, 0x41, 0x6C, 0x67,
    // RemoveAlg
    0x52, 0x65, 0x6D, 0x6F, 0x76, 0x65, 0x41, 0x6C, 0x67,
    // UUID
    0x55, 0x55, 0x49, 0x44,
    // LON - Licence ordinary number
    0x4C, 0x4F, 0x4E,
    // Type
    0x54, 0x79, 0x70, 0x65,
    // NewSWA
    0x4E, 0x65, 0x77, 0x53, 0x57, 0x41,
    // ModifySWA
    0x4D, 0x6F, 0x64, 0x69, 0x66, 0x79, 0x53, 0x57, 0x41,
    // RemoveSWA
    0x52, 0x65, 0x6D, 0x6F, 0x76, 0x65, 0x53, 0x57, 0x41,
    // EKA
    0x45, 0x4B, 0x41,
    // DKA
    0x44, 0x4B, 0x41,
    // EKT
    0x45, 0x4B, 0x54,
    // DKT
    0x44, 0x4B, 0x54,
    // PermitSWA
    0x50, 0x65, 0x72, 0x6D, 0x69, 0x74, 0x53, 0x57, 0x41,
    // UnpermitSWA
    0x55, 0x6E, 0x70, 0x65, 0x72, 0x6D, 0x69, 0x74, 0x53, 0x57, 0x41
  };

  // "MLic"
  private final static byte XML_TAGS_MLic_OFFSET = 1;
  private final static byte XML_TAGS_MLic_LENGTH = 4;
  // "UUIDHWT"
  private final static byte XML_TAGS_UUIDHWT_OFFSET = 5;
  private final static byte XML_TAGS_UUIDHWT_LENGTH = 7;
  // "MLicKey"
  private final static byte XML_TAGS_MLicKey_OFFSET = 12;
  private final static byte XML_TAGS_MLicKey_LENGTH = 7;
  // "ALicKey"
  private final static byte XML_TAGS_ALicKey_OFFSET = 19;
  private final static byte XML_TAGS_ALicKey_LENGTH = 7;
  // "NewAlg"
  private final static byte XML_TAGS_NewAlg_OFFSET = 26;
  private final static byte XML_TAGS_NewAlg_LENGTH = 6;
  // "ModifyAlg"
  private final static byte XML_TAGS_ModifyAlg_OFFSET = 32;
  private final static byte XML_TAGS_ModifyAlg_LENGTH = 9;
  // "RemoveAlg"
  private final static byte XML_TAGS_RemoveAlg_OFFSET = 41;
  private final static byte XML_TAGS_RemoveAlg_LENGTH = 9;
  // "UUID"
  private final static byte XML_TAGS_UUID_OFFSET = 50;
  private final static byte XML_TAGS_UUID_LENGTH = 4;
  // "LON"
  private final static byte XML_TAGS_LON_OFFSET = 54;
  private final static byte XML_TAGS_LON_LENGTH = 3;
  // "Type"
  private final static byte XML_TAGS_Type_OFFSET = 57;
  private final static byte XML_TAGS_Type_LENGTH = 4;
  // "NewSWA"
  private final static byte XML_TAGS_NewSWA_OFFSET = 61;
  private final static byte XML_TAGS_NewSWA_LENGTH = 6;
  // "ModifySWA"
  private final static byte XML_TAGS_ModifySWA_OFFSET = 67;
  private final static byte XML_TAGS_ModifySWA_LENGTH = 9;
  // "RemoveSWA"
  private final static byte XML_TAGS_RemoveSWA_OFFSET = 76;
  private final static byte XML_TAGS_RemoveSWA_LENGTH = 9;
  // "EKA"
  private final static byte XML_TAGS_EKA_OFFSET = 85;
  private final static byte XML_TAGS_EKA_LENGTH = 3;
  // "DKA"
  private final static byte XML_TAGS_DKA_OFFSET = 88;
  private final static byte XML_TAGS_DKA_LENGTH = 3;
  // "EKT"
  private final static short XML_TAGS_EKT_OFFSET = 91;
  private final static byte XML_TAGS_EKT_LENGTH = 3;
  // "DKT"
  private final static short XML_TAGS_DKT_OFFSET = 94;
  private final static byte XML_TAGS_DKT_LENGTH = 3;
  // "PermitSWA"
  private final static short XML_TAGS_PermitSWA_OFFSET = 97;
  private final static byte XML_TAGS_PermitSWA_LENGTH = 9;
  // "UnpermitSWA"
  private final static short XML_TAGS_UnpermitSWA_OFFSET = 106;
  private final static byte XML_TAGS_UnpermitSWA_LENGTH = 11;


  //
  //  TAG ELEMENTS STACK VALUES
  //
  private final static byte TAG_MASTERLICENCE      = (byte) 0x01;
  private final static byte TAG_UUIDHWT            = (byte) 0x02;
  private final static byte TAG_MASTERLICENCEKEY   = (byte) 0x03;
  private final static byte TAG_ALGLICENCEKEY      = (byte) 0x04;
  private final static byte TAG_NEWALG             = (byte) 0x05;
  private final static byte TAG_MODIFYALG          = (byte) 0x06;
  private final static byte TAG_REMOVEALG          = (byte) 0x07;
  private final static byte TAG_UUID               = (byte) 0x08;
  private final static byte TAG_LICORDNUM          = (byte) 0x09;
  private final static byte TAG_TYPE               = (byte) 0x0A;
  private final static byte TAG_NEWSWA             = (byte) 0x0B;
  private final static byte TAG_MODIFYSWA          = (byte) 0x0C;
  private final static byte TAG_REMOVESWA          = (byte) 0x0D;
  private final static byte TAG_EKEYAUTH           = (byte) 0x0E;
  private final static byte TAG_DKEYAUTH           = (byte) 0x0F;
  private final static byte TAG_EKEYTRANSP         = (byte) 0x10;
  private final static byte TAG_DKEYTRANSP         = (byte) 0x11;
  private final static byte TAG_PERMITSWA          = (byte) 0x12;
  private final static byte TAG_UNPERMITSWA        = (byte) 0x13;

  private final static byte TAG_NOTAG              = (byte) 0xFF;

  //
  //  AVAILABLE ALGORITHMS IDENTIFICATORS
  //
  private final static byte ALGTYPEGAUSSEASTER      = (byte) 0x01;


  //
  //  PUBLIC ATTRIBUTES
  //
  // LIST OF ALL AVAILABLE PROFILES
  public SecureAlgProfile     m_secureAlgProfiles[];
  // LIST OF ALL AVAILABLE SWA PROFILES
  public SWAProfile           m_SWAProfiles[];
  // UUID OF PARENT HWT
  public byte                 m_uuidHWT[];
  // LAST USED LICENCE ID, NEW MUST BE HIGHER
  public short                m_lastUsedLicenceOrdNum;
  // PARSED MASTER LICENCE KEY
  public byte                 m_masterLicenceKey[] = null;
  // PARSED ALG LICENCE KEY
  public byte                 m_algLicenceKey[] = null;
  // AFTER PARSING INDICATE, IF PARSED VALUES CAN BE UPDATED
  public boolean              m_correctLicence = false;
  public boolean              m_outOfOrderLicence = false;

  private boolean             m_processLicence = false;
  private byte                m_algType = 0;
  private byte[]              m_UUIDHWT = null;
  private byte[]              m_UUID = null;
  private byte[]              m_encryptKeyAuth = null;
  private byte[]              m_decryptKeyAuth = null;
  private byte[]              m_encryptKeyTransp = null;
  private byte[]              m_decryptKeyTransp = null;
  private byte[]              m_permitSWA = null;
  private byte[]              m_unpermitSWA = null;
  private short               m_licenceOrdNum = 0;
  private Stack               m_tagStack = new Stack();

  //
  //  METHODS
  //
  public XMLMasterLicenceParseHandler() {}

  // NOTE: popTag & topTag CANNOT BE PLACED INTO XMLParseHandler
  // BECAUSE JCRE SOMETIMES SUCKS WHEN USING m_tagStack MEMBER
  /**
   * Pop and return topmost tag from stack
   */
  protected byte popTag() {
    if (m_tagStack.empty()) return TAG_NOTAG;
    else return m_tagStack.pop();
  }

  /**
   * Return topmost tag from stack
   */
  protected byte topTag() {
    if (m_tagStack.empty()) return TAG_NOTAG;
    else return m_tagStack.top();
  }


  public void startDocument() {
    m_tagStack.clear();
    m_masterLicenceKey = null;
    m_algLicenceKey = null;
    m_algType = 0;
    m_licenceOrdNum = 0;
    m_UUID = null;
    m_encryptKeyAuth = null;
    m_decryptKeyAuth = null;
    m_encryptKeyTransp = null;
    m_decryptKeyTransp = null;
    m_permitSWA = null;
    m_unpermitSWA = null;
    m_correctLicence = false;
    m_processLicence = false;
  }
  public void endDocument() {
    if (m_processLicence) m_correctLicence = true;
  }

  public void startElement(CardString tag, CardString attributName, CardString attributValue, boolean newElement) throws ISOException {
    //
    // ELEMENT MLic
    //
    if (tag.compare(XML_TAGS, XML_TAGS_MLic_OFFSET, XML_TAGS_MLic_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_MASTERLICENCE);
      return;
    }

    //
    // ELEMENT MLicKey
    //
    if (tag.compare(XML_TAGS, XML_TAGS_MLicKey_OFFSET, XML_TAGS_MLicKey_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_MASTERLICENCEKEY);
      return;
    }

    //
    // ELEMENT ALicKey
    //
    if (tag.compare(XML_TAGS, XML_TAGS_ALicKey_OFFSET, XML_TAGS_ALicKey_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_ALGLICENCEKEY);
      return;
    }
    //
    // ELEMENT UUIDHWT
    //
    if (tag.compare(XML_TAGS, XML_TAGS_UUIDHWT_OFFSET, XML_TAGS_UUIDHWT_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_UUIDHWT);
      return;
    }
    //
    // ELEMENT UUID
    //
    if (tag.compare(XML_TAGS, XML_TAGS_UUID_OFFSET, XML_TAGS_UUID_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_UUID);
      return;
    }
    //
    // ELEMENT NewAlg
    //
    if (tag.compare(XML_TAGS, XML_TAGS_NewAlg_OFFSET, XML_TAGS_NewAlg_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) {
        m_tagStack.push(TAG_NEWALG);
        m_UUID = null;
        m_algType = 0;
      }
      return;
    }
    //
    // ELEMENT ModifyAlg
    //
    if (tag.compare(XML_TAGS, XML_TAGS_ModifyAlg_OFFSET, XML_TAGS_ModifyAlg_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) {
        m_tagStack.push(TAG_MODIFYALG);
        m_UUID = null;
        m_permitSWA = null;
        m_unpermitSWA = null;
      }
      return;
    }
    //
    // ELEMENT RemoveAlg
    //
    if (tag.compare(XML_TAGS, XML_TAGS_RemoveAlg_OFFSET, XML_TAGS_RemoveAlg_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) {
        m_tagStack.push(TAG_REMOVEALG);
        m_UUID = null;
        m_algType = 0;
      }
      return;
    }
    //
    // ELEMENT LOC
    //
    if (tag.compare(XML_TAGS, XML_TAGS_LON_OFFSET, XML_TAGS_LON_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_LICORDNUM);
      return;
    }
    //
    // ELEMENT Type
    //
    if (tag.compare(XML_TAGS, XML_TAGS_Type_OFFSET, XML_TAGS_Type_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_TYPE);
      return;
    }
    //
    // ELEMENT NewSWA
    //
    if (tag.compare(XML_TAGS, XML_TAGS_NewSWA_OFFSET, XML_TAGS_NewSWA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_NEWSWA);
      return;
    }
    //
    // ELEMENT ModifySWA
    //
    if (tag.compare(XML_TAGS, XML_TAGS_ModifySWA_OFFSET, XML_TAGS_ModifySWA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_MODIFYSWA);
      return;
    }
    //
    // ELEMENT RemoveSWA
    //
    if (tag.compare(XML_TAGS, XML_TAGS_RemoveSWA_OFFSET, XML_TAGS_RemoveSWA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_REMOVESWA);
      return;
    }
    //
    // ELEMENT EKeyAuth
    //
    if (tag.compare(XML_TAGS, XML_TAGS_EKA_OFFSET, XML_TAGS_EKA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_EKEYAUTH);
      return;
    }
    //
    // ELEMENT DKA
    //
    if (tag.compare(XML_TAGS, XML_TAGS_DKA_OFFSET, XML_TAGS_DKA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_DKEYAUTH);
      return;
    }
    //
    // ELEMENT EKT
    //
    if (tag.compare(XML_TAGS, XML_TAGS_EKT_OFFSET, XML_TAGS_EKT_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_EKEYTRANSP);
      return;
    }
    //
    // ELEMENT DKT
    //
    if (tag.compare(XML_TAGS, XML_TAGS_DKT_OFFSET, XML_TAGS_DKT_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_DKEYTRANSP);
      return;
    }
    //
    // ELEMENT PermitSWA
    //
    if (tag.compare(XML_TAGS, XML_TAGS_PermitSWA_OFFSET, XML_TAGS_PermitSWA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_PERMITSWA);
      return;
    }
    //
    // ELEMENT UnpermitSWA
    //
    if (tag.compare(XML_TAGS, XML_TAGS_UnpermitSWA_OFFSET, XML_TAGS_UnpermitSWA_LENGTH) == 0) {
      // PUSH ON STACK
      if (newElement) m_tagStack.push(TAG_UNPERMITSWA);
      return;
    }

    // IF REACH THIS LINE, THAN UNEXPECTED ELEMNT WAS FOUND
    ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
  }

  public void startElementFinal(CardString tag) {}

  public void endElement(CardString elem) throws ISOException{
    //
    // ELEMENT MLic
    //
    if (elem.compare(XML_TAGS, XML_TAGS_MLic_OFFSET, XML_TAGS_MLic_LENGTH) == 0) {
      if (topTag() == TAG_MASTERLICENCE) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT MLicKey
    //
    if (elem.compare(XML_TAGS, XML_TAGS_MLicKey_OFFSET, XML_TAGS_MLicKey_LENGTH) == 0) {
      if (topTag() == TAG_MASTERLICENCEKEY) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT ALicKey
    //
    if (elem.compare(XML_TAGS, XML_TAGS_ALicKey_OFFSET, XML_TAGS_ALicKey_LENGTH) == 0) {
      if (topTag() == TAG_ALGLICENCEKEY) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT UUIDHWT
    //
    if (elem.compare(XML_TAGS, XML_TAGS_UUIDHWT_OFFSET, XML_TAGS_UUIDHWT_LENGTH) == 0) {
      if (topTag() == TAG_UUIDHWT) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT UUID
    //
    if (elem.compare(XML_TAGS, XML_TAGS_UUID_OFFSET, XML_TAGS_UUID_LENGTH) == 0) {
      if (topTag() == TAG_UUID) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT LOC
    //
    if (elem.compare(XML_TAGS, XML_TAGS_LON_OFFSET, XML_TAGS_LON_LENGTH) == 0) {
      if (topTag() == TAG_LICORDNUM) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT Type
    //
    if (elem.compare(XML_TAGS, XML_TAGS_Type_OFFSET, XML_TAGS_Type_LENGTH) == 0) {
      if (topTag() == TAG_TYPE) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT NewAlg
    //
    if (elem.compare(XML_TAGS, XML_TAGS_NewAlg_OFFSET, XML_TAGS_NewAlg_LENGTH) == 0) {
      if (topTag() == TAG_NEWALG) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      // CREATE NEW ALGORITHM
      if (CanBeLicenceProcessed()) {
        boolean   algCreated = false;
        // CHECK UUID
        if ((m_UUID == null) || (m_UUID.length != SecureAlgManager.ALGORITHM_UUID_LENGTH)) ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_BAD_LENGTH);

        if (SecureAlgManager.IsAlgUUIDUnique(m_secureAlgProfiles, m_UUID)) {
          for (byte i = 0; i < m_secureAlgProfiles.length; i++) {
            // FIND FIRST FREE POSITION
            if (m_secureAlgProfiles[i] == null) {
              // CREATE NEW ALGORITH BASED ON SPECIFIED TYPE
              switch (m_algType) {
                case ALGTYPEGAUSSEASTER: {
                  SecureAlgProfile profile = new GaussEaster(m_UUID, (short) 0, (short) m_UUID.length);

                  algCreated = true;
                  m_secureAlgProfiles[i] = profile;
                  break;
                }
                default: {
                  ISOException.throwIt(SecureAlgManager.SW_ALG_TYPE_UNKNOWN);
                  break;
                }
              }
            }
            if (algCreated) break;
          }
          if (!algCreated) ISOException.throwIt(SecureAlgManager.SW_ALG_CONTAINER_FULL);
        }
        else ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_NOT_UNIQUE);
      }

      // CLEAR USED ARRAYS
      m_UUID = null;

      return;
    }

    //
    // ELEMENT ModifyAlg
    //
    if (elem.compare(XML_TAGS, XML_TAGS_ModifyAlg_OFFSET, XML_TAGS_ModifyAlg_LENGTH) == 0) {
      if (topTag() == TAG_MODIFYALG) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    //  ELEMENT RemoveAlg
    //
    if (elem.compare(XML_TAGS, XML_TAGS_RemoveAlg_OFFSET, XML_TAGS_RemoveAlg_LENGTH) == 0) {
      if (topTag() == TAG_REMOVEALG) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      // REMOVE ALGORITHM
      if (CanBeLicenceProcessed()) {
        if (m_UUID != null) {
          boolean algRemoved = false;
          for (byte i = 0; i < m_secureAlgProfiles.length; i++) {
            if (m_secureAlgProfiles[i] != null) {
              if (m_secureAlgProfiles[i].IsAlgorithm(m_UUID, (short) 0, (short) m_UUID.length)) {
                // REMOVE TARGET ALGORITHM FROM ALL SWA WITH THIS ALGORITHM SET AS USAGE PERMITED
                SecureAlgManager.RemoveAllPermitedAlgsByUUID(m_SWAProfiles, m_UUID, (short) 0, (short) m_UUID.length);

                // REMOVE TARGET ALGORITHM
                m_secureAlgProfiles[i] = null;
                algRemoved = true;
                break;
              }
            }
          }

          if (!algRemoved) ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_NOT_FOUND);
        }
        else ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_BAD_LENGTH);
      }
      // CLEAR USED ARRAYS
      m_UUID = null;

      return;
    }

    //
    // ELEMENT NewSWA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_NewSWA_OFFSET, XML_TAGS_NewSWA_LENGTH) == 0) {
      if (topTag() == TAG_NEWSWA) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      // ADD NEW SWA PROFILE
      // CHECK UUID
      if ((m_UUID == null) || (m_UUID.length != SecureAlgManager.SWA_UUID_LENGTH)) ISOException.throwIt(SecureAlgManager.SW_SWA_UUID_BAD_LENGTH);

      if (SecureAlgManager.IsSWAUUIDUnique(m_SWAProfiles, m_UUID)) {
        // CHECK REQUIRED KEYS
        if ((m_encryptKeyAuth != null) && (m_encryptKeyAuth.length == SecureAlgManager.AUTH_KEY_LENGTH) &&
            (m_decryptKeyAuth != null) && (m_decryptKeyAuth.length == SecureAlgManager.AUTH_KEY_LENGTH) &&
            (m_encryptKeyTransp != null) && (m_encryptKeyTransp.length == SecureAlgManager.TRANSPORT_KEY_LENGTH) &&
            (m_decryptKeyTransp != null) && (m_decryptKeyTransp.length == SecureAlgManager.TRANSPORT_KEY_LENGTH)) {

          // FIND FIRST FREE POSITION
          boolean swaCreated = false;
          for (byte i = 0; i < m_SWAProfiles.length; i++) {
            if (m_SWAProfiles[i] == null) {
              // CREATE NEW SWA PROFILE
              SWAProfile   SWAProfile = new SWAProfile();
              Util.arrayCopyNonAtomic(m_UUID, (short) 0, SWAProfile.m_uuidSWA, (short) 0, (short) m_UUID.length);
              Util.arrayCopyNonAtomic(m_encryptKeyAuth, (short) 0, SWAProfile.m_encryptKeyAuth, (short) 0, (short) m_encryptKeyAuth.length);
              Util.arrayCopyNonAtomic(m_decryptKeyAuth, (short) 0, SWAProfile.m_decryptKeyAuth, (short) 0, (short) m_decryptKeyAuth.length);
              Util.arrayCopyNonAtomic(m_encryptKeyTransp, (short) 0, SWAProfile.m_encryptKeyTransp, (short) 0, (short) m_encryptKeyTransp.length);
              Util.arrayCopyNonAtomic(m_decryptKeyTransp, (short) 0, SWAProfile.m_decryptKeyTransp, (short) 0, (short) m_decryptKeyTransp.length);
              swaCreated = true;
              m_SWAProfiles[i] = SWAProfile;
              break;
            }
          }
          if (!swaCreated) ISOException.throwIt(SecureAlgManager.SW_SWA_CONTAINER_FULL);
        }
        else ISOException.throwIt(SecureAlgManager.SW_BAD_KEY);
      }
      else ISOException.throwIt(SecureAlgManager.SW_SWA_UUID_NOT_UNIQUE);

      // CLEAR USED ARRAYS
      ClearDataArrays();
      return;
    }

    //
    // ELEMENT ModifySWA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_ModifySWA_OFFSET, XML_TAGS_ModifySWA_LENGTH) == 0) {
      if (topTag() == TAG_MODIFYSWA) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      // CHECK EXPECTED KEYS
      if (((m_encryptKeyAuth != null) && (m_encryptKeyAuth.length != SecureAlgManager.AUTH_KEY_LENGTH)) ||
          ((m_decryptKeyAuth != null) && (m_decryptKeyAuth.length != SecureAlgManager.AUTH_KEY_LENGTH)) ||
          ((m_encryptKeyTransp != null) && (m_encryptKeyTransp.length != SecureAlgManager.TRANSPORT_KEY_LENGTH)) ||
          ((m_decryptKeyTransp != null) && (m_decryptKeyTransp.length != SecureAlgManager.TRANSPORT_KEY_LENGTH))) {

          // AT LEAST ONE KEY IS BAD
          ISOException.throwIt(SecureAlgManager.SW_BAD_KEY);
      }
      else {
        // FIND TARGET SWA PROFILE
        boolean     swaModified = false;
        SWAProfile  actualSWA;
        if ((actualSWA = SecureAlgManager.GetSWAByUUID(m_SWAProfiles, m_UUID, (short) 0, (short) m_UUID.length)) != null) {
          // MODIFY SWA PROFILE
          if (m_encryptKeyAuth != null) Util.arrayCopyNonAtomic(m_encryptKeyAuth, (short) 0, actualSWA.m_encryptKeyAuth, (short) 0, (short) m_encryptKeyAuth.length);
          if (m_decryptKeyAuth != null) Util.arrayCopyNonAtomic(m_decryptKeyAuth, (short) 0, actualSWA.m_decryptKeyAuth, (short) 0, (short) m_decryptKeyAuth.length);
          if (m_encryptKeyTransp != null) Util.arrayCopyNonAtomic(m_encryptKeyTransp, (short) 0, actualSWA.m_encryptKeyTransp, (short) 0, (short) m_encryptKeyTransp.length);
          if (m_decryptKeyTransp != null) Util.arrayCopyNonAtomic(m_decryptKeyTransp, (short) 0, actualSWA.m_decryptKeyTransp, (short) 0, (short) m_decryptKeyTransp.length);
          swaModified = true;
        }
        if (!swaModified) ISOException.throwIt(SecureAlgManager.SW_SWA_UUID_BAD);
      }

      // CLEAR USED ARRAYS
      ClearDataArrays();

      return;
    }
    //
    // ELEMENT RemoveSWA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_RemoveSWA_OFFSET, XML_TAGS_RemoveSWA_LENGTH) == 0) {
      if (topTag() == TAG_REMOVESWA) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      // FIND TARGET SWA PROFILE
      boolean swaRemoved = false;
      for (byte i = 0; i < m_SWAProfiles.length; i++) {
        if ((m_SWAProfiles[i] != null) && (m_SWAProfiles[i].IsSWAUUID(m_UUID, (short) 0, (short) m_UUID.length))) {
          m_SWAProfiles[i].FreeResources();
          m_SWAProfiles[i] = null;
          swaRemoved = true;
        }
      }

      if (!swaRemoved) ISOException.throwIt(SecureAlgManager.SW_SWA_NOT_FOUND);

      // CLEAR USED ARRAYS
      ClearDataArrays();

      return;
    }
    //
    // ELEMENT EKA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_EKA_OFFSET, XML_TAGS_EKA_LENGTH) == 0) {
      if (topTag() == TAG_EKEYAUTH) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    // ELEMENT DKA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_DKA_OFFSET, XML_TAGS_DKA_LENGTH) == 0) {
      if (topTag() == TAG_DKEYAUTH) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    // ELEMENT EKT
    //
    if (elem.compare(XML_TAGS, XML_TAGS_EKT_OFFSET, XML_TAGS_EKT_LENGTH) == 0) {
      if (topTag() == TAG_EKEYTRANSP) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    // ELEMENT DKT
    //
    if (elem.compare(XML_TAGS, XML_TAGS_DKT_OFFSET, XML_TAGS_DKT_LENGTH) == 0) {
      if (topTag() == TAG_DKEYTRANSP) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);
      return;
    }
    //
    // ELEMENT PermitSWA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_PermitSWA_OFFSET, XML_TAGS_PermitSWA_LENGTH) == 0) {
      if (topTag() == TAG_PERMITSWA) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      // ADD PERMITED SWA TO GIVEN ALGORITHM PROFILE
      if (CanBeLicenceProcessed()) {
        if ((m_UUID != null) && (m_permitSWA != null)) {
          // FIND TARGET ALGORITHM PROFILE
          SecureAlgProfile algProfile;
          if ((algProfile = SecureAlgManager.GetAlgorithmByUUID(m_secureAlgProfiles, m_UUID, (short) 0, (short) m_UUID.length)) != null) {
            // GET SWA PROFILE BY m_permitSWA
            SWAProfile swaProfile;
            if ((swaProfile = SecureAlgManager.GetSWAByUUID(m_SWAProfiles, m_permitSWA, (short) 0, (short) m_permitSWA.length)) != null) {
              // PERMIT SWA
              swaProfile.AddPermitedAlg(algProfile);
            }
            else ISOException.throwIt(SecureAlgManager.SW_SWA_UUID_BAD);
          }
          else ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_NOT_FOUND);
        }
        else ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_BAD_LENGTH);
      }

      // CLEAR USED ARRAYS
      m_permitSWA = null;
      return;
    }
    //
    // ELEMENT UnpermitSWA
    //
    if (elem.compare(XML_TAGS, XML_TAGS_UnpermitSWA_OFFSET, XML_TAGS_UnpermitSWA_LENGTH) == 0) {
      if (topTag() == TAG_UNPERMITSWA) popTag();
      else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_ELEM);

      if (CanBeLicenceProcessed()) {
        if ((m_UUID != null) && (m_unpermitSWA != null)) {
          // GET SWA PROFILE BY m_permitSWA
          SWAProfile swaProfile;
          if ((swaProfile = SecureAlgManager.GetSWAByUUID(m_SWAProfiles, m_unpermitSWA, (short) 0, (short) m_unpermitSWA.length)) != null) {
              // REMOVE ALG IF USAGE PERMITED
              swaProfile.RemoveIfPermitedAlg(m_UUID, (short) 0, (short) m_UUID.length);
          }
          else ISOException.throwIt(SecureAlgManager.SW_SWA_UUID_BAD);
        }
        else ISOException.throwIt(SecureAlgManager.SW_ALG_UUID_BAD_LENGTH);
      }

       // CLEAR USED ARRAY
      m_permitSWA = null;

      return;
    }


    //
    //  ENABLE NEXT popTag() COMMAND WHEN 'NON-VALIDATING END OF TAG' IS ALLOWED
    //  NOTE: IF NEXT popTag() IS ALLOWED, THAN CORRECT SEQENCE OF OPEN/CLOSE TAGS IS NOT CONTROLLED EXCEPT TAGS LISTED ABOVE THIS LINE
    //  popTag();
  }

  public void text(CardString text) throws ISOException {
    switch (topTag()) {
      case TAG_MASTERLICENCEKEY: {
        // COPY MASTER LICENCE KEY
        if (CanBeLicenceProcessed()) {
          m_masterLicenceKey = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_masterLicenceKey, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_ALGLICENCEKEY: {
        // COPY ALG LICENCE KEY
        if (CanBeLicenceProcessed()) {
          m_algLicenceKey = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_algLicenceKey, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_UUIDHWT: {
        m_UUIDHWT = new byte[text.m_end];
        Util.arrayCopy(text.m_buffer, (short) 0, m_UUIDHWT, (short) 0, text.m_end);
        break;
      }
      case TAG_UUID: {
        if (CanBeLicenceProcessed()) {
          m_UUID = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_UUID, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_LICORDNUM: {
        m_licenceOrdNum = text.getShortNumber();
        break;
      }
      case TAG_TYPE: {
        if (CanBeLicenceProcessed()) {
          m_algType = (byte) text.getShortNumber();
        }
        break;
      }
      case TAG_EKEYAUTH: {
        if (CanBeLicenceProcessed()) {
          m_encryptKeyAuth = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_encryptKeyAuth, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_DKEYAUTH: {
        if (CanBeLicenceProcessed()) {
          m_decryptKeyAuth = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_decryptKeyAuth, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_EKEYTRANSP: {
        if (CanBeLicenceProcessed()) {
          m_encryptKeyTransp = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_encryptKeyTransp, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_DKEYTRANSP: {
        if (CanBeLicenceProcessed()) {
          m_decryptKeyTransp = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_decryptKeyTransp, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_PERMITSWA: {
        if (CanBeLicenceProcessed()) {
          m_permitSWA = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_permitSWA, (short) 0, text.m_end);
        }
        break;
      }
      case TAG_UNPERMITSWA: {
        if (CanBeLicenceProcessed()) {
          m_unpermitSWA = new byte[text.m_end];
          Util.arrayCopy(text.m_buffer, (short) 0, m_unpermitSWA, (short) 0, text.m_end);
        }
        break;
      }
      default: {
        // NO ACTION
      }
    }
  }


  private boolean CanBeLicenceProcessed() {
    boolean status = false;
    // CHECK IF LICENCE INFORMATIONS CAN BE PROCESSED
    if (m_processLicence) status = true;
    else {
      if ((m_lastUsedLicenceOrdNum < m_licenceOrdNum) && (m_UUIDHWT != null)) {
        if ((((short) (m_lastUsedLicenceOrdNum + 1)) ==  m_licenceOrdNum) || m_outOfOrderLicence) {
          if (Util.arrayCompare(m_UUIDHWT, (short) 0, m_uuidHWT, (short) 0, (short) m_uuidHWT.length) == 0) {
            // ALLOW PROCESSING LICENCE INFORMATION
            m_processLicence = true;
            m_lastUsedLicenceOrdNum = m_licenceOrdNum;
            status = true;
          }
        }
        else ISOException.throwIt(SecureAlgManager.SW_LICENCE_OUTOFORDER);
      }
      else ISOException.throwIt(SecureAlgManager.SW_LICENCE_OUTDATED);
    }

    return status;
  }


  private void ClearDataArrays() {
    m_encryptKeyAuth = null;
    m_decryptKeyAuth = null;
    m_encryptKeyTransp = null;
    m_decryptKeyTransp = null;
    m_UUID = null;
    m_permitSWA = null;
    m_unpermitSWA = null;
  }
}