/*  XMLParser.java -- simple XML parser

    This is part of SecureFW project, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/dpthesis/index.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

package com.svenda.SecureAlg;
import javacard.framework.*;

public class XMLParser {
  Stack  m_modeStack = new Stack();       // MODE STACK

  private byte popMode() {
    if (m_modeStack.empty()) return PRE;
    else return m_modeStack.pop();
  }

  private final static byte
    TEXT = 1,
    ENTITY = 2,
    OPEN_TAG = 3,
    CLOSE_TAG = 4,
    START_TAG = 5,
    ATTRIBUTE_LVALUE = 6,
    ATTRIBUTE_EQUAL = 9,
    ATTRIBUTE_RVALUE = 10,
    QUOTE = 7,
    IN_TAG = 8,
    SINGLE_TAG = 12,
    COMMENT = 13,
    DONE = 11,
    DOCTYPE = 14,
    PRE = 15,
    CDATA = 16;

  private final byte[] XML_ENTITES = {
//    "\0"| "]]"  | "--" | "!-"  | "![CDATA"                  | "!DOCTYP"                |
      00,  93, 93, 45, 45, 33, 45, 33, 91, 67, 68, 65, 84, 65, 33, 68, 79, 67, 84, 89, 80,
//     "lt"   |  "gt"   | "amp"       |   "quot"          |  "apos"           |
      108, 116, 103, 116, 97, 109, 112, 113, 117, 111, 116, 97, 112, 111, 115
  };

  // "]]"
  private final static byte XMLENT_BB_OFFSET = 1;
  private final static byte XMLENT_BB_LENGTH = 2;

  // "--"
  private final static byte XMLENT_MM_OFFSET = 3;
  private final static byte XMLENT_MM_LENGTH = 2;

  // "!-"
  private final static byte XMLENT_NOTM_OFFSET = 5;
  private final static byte XMLENT_NOTM_LENGTH = 2;

  // "![CDATA"
  private final static byte XMLENT_NOTBCDATA_OFFSET = 7;
  private final static byte XMLENT_NOTBCDATA_LENGTH = 7;

  // "!DOCTYP"
  private final static byte XMLENT_NOTDOCTYP_OFFSET = 14;
  private final static byte XMLENT_NOTDOCTYP_LENGTH = 7;

  // "lt"
  private final static byte XMLENT_lt_OFFSET = 21;
  private final static byte XMLENT_lt_LENGTH = 2;

    // "gt"
  private final static byte XMLENT_gt_OFFSET = 23;
  private final static byte XMLENT_gt_LENGTH = 2;

    // "amp"
  private final static byte XMLENT_amp_OFFSET = 25;
  private final static byte XMLENT_amp_LENGTH = 3;

    // "quot"
  private final static byte XMLENT_quot_OFFSET = 28;
  private final static byte XMLENT_quot_LENGTH = 4;

    // "apos"
  private final static byte XMLENT_apos_OFFSET = 32;
  private final static byte XMLENT_apos_LENGTH = 4;



  public void parse(XMLParseHandler doc, byte[] document, short offset, short documentLength) throws ISOException, Exception {
    CardStringBuffer  sb = new CardStringBuffer();
    CardStringBuffer  etag = new CardStringBuffer();
    CardString        tagName = null;
    CardString        lvalue = null;
    CardString        rvalue = null;
    short             line = 1;
    short             col = 0;
    short             pos = 0;
    short             depth = 0;
    short             quotec = '"';
    boolean           eol = false;
    boolean           newElement = true;
    byte              mode = PRE;
    byte              c = 0;

    doc.startDocument();
    while(pos < documentLength) {
      // ACTUAL CHARACTER
      c = document[(short) (offset + pos)];
      pos++;

      // We need to map \r, \r\n, and \n to \n
      // See XML spec section 2.11
      if(c == '\n' && eol) {
        eol = false;
	continue;
      } else if(eol) {
        eol = false;
      } else if(c == '\n') {
        line++;
	col=0;
      } else if(c == '\r') {
        eol = true;
	c = '\n';
	line++;
	col=0;
      } else {
        col++;
      }

      if (mode == DONE) {
        doc.endDocument();
        return;

      // We are between tags collecting text.
      } else if(mode == TEXT) {
        if(c == '<') {
          m_modeStack.push(mode);
          mode = START_TAG;
          if(sb.length() > 0) {
            doc.text(sb);
            sb.setLength((short) 0);
          }
        }
        else if(c == '&') {
          m_modeStack.push(mode);
          mode = ENTITY;
          etag.setLength((short) 0);
        } else
          sb.append((byte)c);

      // we are processing a closing tag: e.g. </foo>
      } else if(mode == CLOSE_TAG) {
        if(c == '>') {
          mode = popMode();
          depth--;
          if(depth==0)
            mode = DONE;
          doc.endElement(sb);
          sb.setLength((short) 0);
        } else {
          sb.append((byte)c);
        }

      // we are processing CDATA
      } else if(mode == CDATA) {
        if(c == '>'
        && sb.endsWith(XML_ENTITES, XMLENT_BB_OFFSET, XMLENT_BB_LENGTH)) {
          sb.setLength((short) (sb.length()-2));
          doc.text(sb);
          sb.setLength((short) 0);
          mode = popMode();
        } else
          sb.append((byte)c);

      // we are processing a comment.  We are inside
      // the <!-- .... --> looking for the -->.
      } else if(mode == COMMENT) {
        if(c == '>'
        && sb.endsWith(XML_ENTITES, XMLENT_MM_OFFSET, XMLENT_MM_LENGTH)) {
          sb.setLength((short) 0);
          mode = popMode();
        } else
          sb.append((byte)c);

      // We are outside the root tag element
      } else if(mode == PRE) {
        if(c == '<') {
          mode = TEXT;
          m_modeStack.push(mode);
          mode = START_TAG;
        }

      // We are inside one of these <? ... ?>
      // or one of these <!DOCTYPE ... >
      } else if(mode == DOCTYPE) {
        if(c == '>') {
          mode = popMode();
          if(mode == TEXT) mode = PRE;
        }

      // we have just seen a < and
      // are wondering what we are looking at
      // <foo>, </foo>, <!-- ... --->, etc.
      } else if(mode == START_TAG) {
        mode = popMode();
        if(c == '/') {
          m_modeStack.push(mode);
          mode = CLOSE_TAG;
        } else if (c == '?') {
          mode = DOCTYPE;
        } else {
          m_modeStack.push(mode);
          mode = OPEN_TAG;
          tagName = null;
          newElement = true;
          sb.append((byte)c);
        }
      // we are processing an entity, e.g. &lt;, &#187;, etc.
      } else if(mode == ENTITY) {
        if(c == ';') {
          mode = popMode();
          CardString cent = new CardString(etag.getBuffer());
          etag.setLength((short) 0);
          if(cent.compare(XML_ENTITES, XMLENT_lt_OFFSET, XMLENT_lt_LENGTH) == 0)
            sb.append((byte) '<');
          else if(cent.compare(XML_ENTITES, XMLENT_gt_OFFSET, XMLENT_gt_LENGTH) == 0)
            sb.append((byte) '>');
          else if(cent.compare(XML_ENTITES, XMLENT_amp_OFFSET, XMLENT_amp_LENGTH) == 0)
            sb.append((byte) '&');
          else if(cent.compare(XML_ENTITES, XMLENT_quot_OFFSET, XMLENT_quot_LENGTH) == 0)
            sb.append((byte) '"');
          else if(cent.compare(XML_ENTITES, XMLENT_apos_OFFSET, XMLENT_apos_LENGTH) == 0)
            sb.append((byte) '\'');

          // CUSTOM ENTITY DEFINITIONS CAN BE INSERTED HERE

          else ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_INVALID_DATA);
        } else {
          etag.append((byte)c);
        }
      // we have just seen something like this:
      // <foo a="b"/
      // and are looking for the final >.
      } else if (mode == SINGLE_TAG) {
        if (tagName == null) tagName = new CardString(sb.getBuffer());

        if (c != '>') ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_INVALID_DATA);

        doc.startElement(tagName, null, null, newElement);
        doc.startElementFinal(tagName);
        doc.endElement(tagName);
        if(depth==0) {
          doc.endDocument();
          return;
        }

        sb.setLength((short) 0);
        newElement = true;
        tagName = null;
        mode = popMode();

      // we are processing something
      // like this <foo ... >.  It could
      // still be a <!-- ... --> or something.
      } else if(mode == OPEN_TAG) {
        if(c == '>') {
          if(tagName == null)
            tagName = new CardString(sb.getBuffer());
          sb.setLength((short) 0);
          depth++;
          doc.startElement(tagName, lvalue, rvalue, newElement);
          doc.startElementFinal(tagName);
          tagName = null;
          newElement = true;
          mode = popMode();
        } else if(c == '/') {
          mode = SINGLE_TAG;
        } else if(c == '-' && (sb.compare(XML_ENTITES, XMLENT_NOTM_OFFSET, XMLENT_NOTM_LENGTH) == 0)) {
          mode = COMMENT;
        } else if(c == '[' && (sb.compare(XML_ENTITES, XMLENT_NOTBCDATA_OFFSET, XMLENT_NOTBCDATA_LENGTH) == 0)) {
          mode = CDATA;
          sb.setLength((short) 0);
        } else if(c == 'E' && (sb.compare(XML_ENTITES, XMLENT_NOTDOCTYP_OFFSET, XMLENT_NOTDOCTYP_LENGTH) == 0)) {
          sb.setLength((short) 0);
          mode = DOCTYPE;
        } else if(CardString.isWhitespace(c)) {
          tagName = new CardString(sb.getBuffer());
          sb.setLength((short) 0);
          mode = IN_TAG;
        } else {
          sb.append((byte)c);
        }

      // We are processing the quoted right-hand side
      // of an element's attribute.
      } else if(mode == QUOTE) {
        if(c == quotec) {
          rvalue = new CardString(sb.getBuffer());
          sb.setLength((short) 0);
          doc.startElement(tagName, lvalue, rvalue, newElement);
          newElement = false;
          lvalue = rvalue = null;
          mode = IN_TAG;
        // See section the XML spec, section 3.3.3
        // on normalization processing.
        } else if(CardString.isWhitespace(c)) {
          sb.append((byte) ' ');
        } else if(c == '&') {
          m_modeStack.push(mode);
          mode = ENTITY;
          etag.setLength((short) 0);
        } else {
          sb.append((byte)c);
        }

      } else if(mode == ATTRIBUTE_RVALUE) {
        if(c == '"' || c == '\'') {
          quotec = c;
          mode = QUOTE;
        } else if(CardString.isWhitespace(c)) {
          ;
        } else ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_INVALID_DATA);

      } else if(mode == ATTRIBUTE_LVALUE) {
        if(CardString.isWhitespace(c)) {
          lvalue = new CardString(sb.getBuffer());
          sb.setLength((short) 0);
          mode = ATTRIBUTE_EQUAL;
        } else if(c == '=') {
          lvalue = new CardString(sb.getBuffer());
          sb.setLength((short) 0);
          mode = ATTRIBUTE_RVALUE;
        } else {
          sb.append((byte)c);
        }

      } else if(mode == ATTRIBUTE_EQUAL) {
        if(c == '=') {
          mode = ATTRIBUTE_RVALUE;
        } else if(CardString.isWhitespace(c)) {
          ;
        } else ISOException.throwIt(SecureAlgManager.SW_XMLPARSER_INVALID_DATA);

      } else if(mode == IN_TAG) {
        if(c == '>') {
          mode = popMode();
          doc.startElement(tagName, null, null, newElement);
          doc.startElementFinal(tagName);
          depth++;
          tagName = null;
          newElement = true;
        } else if(c == '/') {
          mode = SINGLE_TAG;
        } else if(CardString.isWhitespace(c)) {
          ;
        } else {
          mode = ATTRIBUTE_LVALUE;
          sb.append((byte)c);
        }
      }
    }

    // 'DONE' MODE EXPECTED
    if(mode == DONE)doc.endDocument();
    else ISOException.throwIt(SecureAlgManager.SW_XML_UNEXPECTED_END);
  }
}

